"""Defines the templaters."""

import copy
import importlib
import importlib.util
import logging
import os.path
import pkgutil
import re
import sys
from collections.abc import Iterable, Iterator
from functools import reduce
from typing import (
    Any,
    Callable,
    Optional,
    Union,
    cast,
)

import jinja2.nodes
import jinja2.parser
from jinja2 import (
    Environment,
    FileSystemLoader,
    TemplateError,
    TemplateSyntaxError,
    meta,
)
from jinja2.exceptions import TemplateNotFound, UndefinedError
from jinja2.ext import Extension
from jinja2.sandbox import SandboxedEnvironment

from sqlfluff.core.config import FluffConfig
from sqlfluff.core.errors import SQLFluffUserError, SQLTemplaterError
from sqlfluff.core.formatter import FormatterInterface
from sqlfluff.core.helpers.file import get_encoding
from sqlfluff.core.helpers.slice import is_zero_slice, slice_length
from sqlfluff.core.templaters.base import (
    RawFileSlice,
    TemplatedFile,
    TemplatedFileSlice,
    large_file_check,
)
from sqlfluff.core.templaters.builtins.dbt import DBT_BUILTINS, DbtMacroWrapper
from sqlfluff.core.templaters.python import PythonTemplater
from sqlfluff.core.templaters.slicers.tracer import JinjaAnalyzer, JinjaTrace

# Instantiate the templater logger
templater_logger = logging.getLogger("sqlfluff.templater")


class UndefinedRecorder:
    """Similar to jinja2.StrictUndefined, but remembers, not fails."""

    # Tell Jinja this object is safe to call and does not alter data.
    # https://jinja.palletsprojects.com/en/3.0.x/sandbox/#jinja2.sandbox.SandboxedEnvironment.is_safe_callable
    unsafe_callable = False
    alters_data = False

    def __init__(self, name: str, undefined_set: set[str]) -> None:
        self.name = name
        # Reference to undefined set to modify, it is assumed that the
        # calling code keeps a reference to this variable to they can
        # continue to access it after modification by this class.
        self.undefined_set = undefined_set

    def __str__(self) -> str:
        """Treat undefined vars as empty, but remember for later."""
        self.undefined_set.add(self.name)
        return ""

    def __getattr__(self, item: str) -> "UndefinedRecorder":
        """Don't fail when called, remember instead."""
        self.undefined_set.add(self.name)
        return UndefinedRecorder(f"{self.name}.{item}", self.undefined_set)

    def __getitem__(self, item: str) -> "UndefinedRecorder":
        """Don't fail when called, remember instead."""
        self.undefined_set.add(self.name)
        return UndefinedRecorder(f"{self.name}.{item}", self.undefined_set)

    def __call__(self, *args: Any, **kwargs: Any) -> "UndefinedRecorder":
        """Don't fail when called unlike parent class."""
        return UndefinedRecorder(f"{self.name}()", self.undefined_set)

    def __iter__(self) -> Iterator["UndefinedRecorder"]:
        """Don't fail when iterated, remember instead."""
        self.undefined_set.add(self.name)
        yield UndefinedRecorder(f"iter({self.name})", self.undefined_set)


class JinjaTemplater(PythonTemplater):
    """A templater using the jinja2 library.

    See: https://jinja.palletsprojects.com/
    """

    name = "jinja"

    class Libraries:
        """Mock namespace for user-defined Jinja library."""

        pass

    @staticmethod
    def _is_trim_tag(raw_slice: RawFileSlice) -> bool:
        """Return whether a raw slice is a Jinja tag with whitespace trimming."""
        return raw_slice.slice_type.startswith("block") and (
            raw_slice.raw.startswith("{%-") or raw_slice.raw.endswith("-%}")
        )

    @classmethod
    def _is_unreached_whitespace_adjacent_to_trim_tag(
        cls, raw_sliced: list[RawFileSlice], idx: int
    ) -> bool:
        """Return whether an unreached literal is only trim-adjacent whitespace."""
        raw_slice = raw_sliced[idx]
        if raw_slice.slice_type != "literal" or not raw_slice.raw.isspace():
            return False

        previous_slice = raw_sliced[idx - 1] if idx > 0 else None
        next_slice = raw_sliced[idx + 1] if idx + 1 < len(raw_sliced) else None
        return bool(
            (previous_slice and cls._is_trim_tag(previous_slice))
            or (next_slice and cls._is_trim_tag(next_slice))
        )

    @staticmethod
    def _extract_macros_from_template(
        template: str, env: Environment, ctx: dict[str, Any]
    ) -> dict[str, DbtMacroWrapper]:
        """Take a template string and extract any macros from it.

        Lovingly inspired by http://codyaray.com/2015/05/auto-load-jinja2-macros

        Raises:
            TemplateSyntaxError: If the macro we try to load has invalid
                syntax. We assume that outer functions will catch this
                exception and handle it appropriately.
        """
        from jinja2.runtime import Macro

        # Iterate through keys exported from the loaded template string
        context: dict[str, DbtMacroWrapper] = {}
        # NOTE: `env.from_string()` will raise TemplateSyntaxError if `template`
        # is invalid.
        macro_template = env.from_string(template, globals=ctx)

        # This is kind of low level and hacky but it works
        try:
            for k in macro_template.module.__dict__:
                attr = getattr(macro_template.module, k)
                # Is it a macro? If so install it at the name of the macro
                if isinstance(attr, Macro):
                    context[k] = DbtMacroWrapper(attr)
        except UndefinedError:
            # This occurs if any file in the macro path references an
            # undefined Jinja variable. It's safe to ignore this. Any
            # meaningful issues will surface later at linting time.
            pass
        # Return the context
        return context

    @classmethod
    def _extract_macros_from_path(
        cls,
        path: list[str],
        env: Environment,
        ctx: dict[str, Any],
        exclude_paths: Optional[list[str]] = None,
        config_encoding: str = "autodetect",
    ) -> dict[str, DbtMacroWrapper]:
        """Take a path and extract macros from it.

        Args:
            path (list[str]): A list of paths.
            env (Environment): The environment object.
            ctx (Dict): The context dictionary.
            exclude_paths (Optional[[List][str]]): A list of paths to exclude
            config_encoding: The configured encoding or autodetect.

        Returns:
            dict: A dictionary containing the extracted macros.

        Raises:
            ValueError: If a path does not exist.
            SQLTemplaterError: If there is an error in the Jinja macro file.
        """
        macro_ctx: dict[str, DbtMacroWrapper] = {}
        for path_entry in path:
            # Does it exist? It should as this check was done on config load.
            if not os.path.exists(path_entry):
                raise ValueError(f"Path does not exist: {path_entry}")

            if os.path.isfile(path_entry):
                if exclude_paths:
                    if cls._exclude_macros(
                        macro_path=path_entry, exclude_macros_path=exclude_paths
                    ):
                        continue
                # It's a file. Extract macros from it.
                encoding = get_encoding(
                    fname=path_entry, config_encoding=config_encoding
                )
                with open(path_entry, encoding=encoding) as opened_file:
                    template = opened_file.read()
                # Update the context with macros from the file.
                try:
                    macro_ctx.update(
                        cls._extract_macros_from_template(template, env=env, ctx=ctx)
                    )
                except TemplateSyntaxError as err:
                    raise SQLTemplaterError(
                        f"Error in Jinja macro file {os.path.relpath(path_entry)}: "
                        f"{err.message}",
                        line_no=err.lineno,
                        line_pos=1,
                    ) from err
            else:
                # It's a directory. Iterate through files in it and extract from them.
                for dirpath, _, files in os.walk(path_entry):
                    for fname in files:
                        if fname.endswith(".sql"):
                            macro_ctx.update(
                                cls._extract_macros_from_path(
                                    [os.path.join(dirpath, fname)],
                                    env=env,
                                    ctx=ctx,
                                    exclude_paths=exclude_paths,
                                    config_encoding=config_encoding,
                                )
                            )
        return macro_ctx

    def _extract_macros_from_config(
        self, config: FluffConfig, env: Environment, ctx: dict[str, Any]
    ) -> dict[str, DbtMacroWrapper]:
        """Take a config and load any macros from it.

        Args:
            config: The config to extract macros from.
            env: The environment.
            ctx: The context.

        Returns:
            dict: A dictionary containing the extracted macros.
        """
        if config:
            loaded_context = (
                config.get_section((self.templater_selector, self.name, "macros")) or {}
            )
        else:  # pragma: no cover TODO?
            loaded_context = {}

        # Iterate to load macros
        macro_ctx: dict[str, DbtMacroWrapper] = {}
        for value in loaded_context.values():
            try:
                macro_ctx.update(
                    self._extract_macros_from_template(value, env=env, ctx=ctx)
                )
            except TemplateSyntaxError as err:
                raise SQLFluffUserError(
                    f"Error loading user provided macro:\n`{value}`\n> {err}."
                )
        return macro_ctx

    def _extract_macros(
        self, config: FluffConfig, env: Environment, ctx: dict[str, Any]
    ) -> dict[str, DbtMacroWrapper]:
        """Load macros directly in a config and from load_macros_from_path.

        Args:
            config: The config to extract macros from.
            env: The environment.
            ctx: The context.

        Returns:
            dict: A dictionary containing the extracted macros.
        """
        macro_ctx: dict[str, DbtMacroWrapper] = {}

        macros_path = self._get_macros_path(config, "load_macros_from_path")
        exclude_macros_path = self._get_macros_path(config, "exclude_macros_from_path")
        if macros_path:
            config_encoding: str = config.get("encoding", default="autodetect")
            macro_ctx.update(
                self._extract_macros_from_path(
                    macros_path,
                    env=env,
                    ctx=ctx,
                    exclude_paths=exclude_macros_path,
                    config_encoding=config_encoding,
                )
            )

        # Load config macros, these will take precedence over macros from the path
        macro_ctx.update(
            self._extract_macros_from_config(config=config, env=env, ctx=ctx)
        )

        return macro_ctx

    def _extract_libraries_from_config(self, config: FluffConfig) -> dict[str, Any]:
        """Extracts libraries from the given configuration.

        This function iterates over the modules in the library path and
        imports them dynamically. The imported modules are then added to a 'Libraries'
        object, which is returned as a dictionary excluding magic methods.

        Args:
            config: The configuration object.

        Returns:
            dict: A dictionary containing the extracted libraries.
        """
        # If a more global library_path is set, let that take precedence.
        library_path = config.get("library_path") or config.get_section(
            (self.templater_selector, self.name, "library_path")
        )
        if not library_path:
            return {}

        libraries = JinjaTemplater.Libraries()

        # If library_path has __init__.py we parse it as one module, else we parse it
        # a set of modules
        is_library_module = os.path.exists(os.path.join(library_path, "__init__.py"))
        library_module_name = os.path.basename(library_path)

        # Need to go one level up to parse as a module correctly
        walk_path = (
            os.path.join(library_path, "..") if is_library_module else library_path
        )

        for module_finder, module_name, _ in pkgutil.walk_packages([walk_path]):
            # skip other modules that can be near module_dir
            if is_library_module and not module_name.startswith(library_module_name):
                continue

            # import_module is deprecated as of python 3.4. This follows roughly
            # the guidance of the python docs:
            # https://docs.python.org/3/library/importlib.html#approximating-importlib-import-module
            spec = module_finder.find_spec(module_name, None)
            assert spec, (
                f"Module {module_name} failed to be found despite being listed."
            )
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            assert spec.loader, f"Module {module_name} missing expected loader."
            spec.loader.exec_module(module)

            if "." in module_name:  # nested modules have `.` in module_name
                *module_path, last_module_name = module_name.split(".")
                # find parent module recursively
                parent_module = reduce(
                    lambda res, path_part: getattr(res, path_part),
                    module_path,
                    libraries,
                )

                # set attribute on module object to make jinja working correctly
                setattr(parent_module, last_module_name, module)
            else:
                # set attr on `libraries` obj to make it work in jinja nicely
                setattr(libraries, module_name, module)

        if is_library_module:
            # when library is module we have one more root module in hierarchy and we
            # remove it
            libraries = getattr(libraries, library_module_name)

        # remove magic methods from result
        return {k: v for k, v in libraries.__dict__.items() if not k.startswith("__")}

    @classmethod
    def _crawl_tree(
        cls, tree: jinja2.nodes.Node, variable_names: set[str], raw: str
    ) -> Iterator[SQLTemplaterError]:
        """Crawl the tree looking for occurrences of the undeclared values."""
        # First iterate through children
        for elem in tree.iter_child_nodes():
            yield from cls._crawl_tree(elem, variable_names, raw)
        # Then assess self
        if (
            isinstance(tree, jinja2.nodes.Name)
            and getattr(tree, "name") in variable_names
        ):
            line_no: int = getattr(tree, "lineno")
            tree_name: str = getattr(tree, "name")
            line = raw.split("\n")[line_no - 1]
            pos = line.index(tree_name) + 1
            yield SQLTemplaterError(
                f"Undefined jinja template variable: {tree_name!r}",
                line_no=line_no,
                line_pos=pos,
            )

    def _get_jinja_env(self, config: Optional[FluffConfig] = None) -> Environment:
        """Get a properly configured jinja environment.

        This method returns a properly configured jinja environment. It
        first checks if the 'ignore' key is present in the config dictionary and
        if it contains the value 'templating'. If so, it creates a subclass of
        FileSystemLoader called SafeFileSystemLoader that overrides the
        get_source method to handle missing templates when templating is ignored.
        If 'ignore' is not present or does not contain 'templating', it uses the
        regular FileSystemLoader. It then sets the extensions to ['jinja2.ext.do']
        and adds the DBTTestExtension if the _apply_dbt_builtins method returns
        True. Finally, it returns a SandboxedEnvironment object with the
        specified settings.

        Args:
            config (dict, optional): A dictionary containing configuration settings.

        Returns:
            jinja2.Environment: A properly configured jinja environment.
        """
        loader: Optional[FileSystemLoader]
        macros_path = self._get_macros_path(config, "load_macros_from_path")
        loader_search_path = self._get_loader_search_path(config)
        final_search_path = (loader_search_path or []) + (macros_path or [])

        ignore_templating = config and "templating" in config.get("ignore")
        if ignore_templating:

            class SafeFileSystemLoader(FileSystemLoader):
                def get_source(
                    self, environment: Environment, name: str
                ) -> tuple[str, str, Callable[..., Any]]:
                    try:
                        if not isinstance(name, DummyUndefined):
                            return super().get_source(environment, name)
                        raise TemplateNotFound(str(name))
                    except TemplateNotFound:
                        # When ignore=templating is set, treat missing files
                        # or attempts to load an "Undefined" file as the first
                        # 'base' part of the name / filename rather than failing.
                        templater_logger.debug(
                            "Providing dummy contents for Jinja macro file: %s", name
                        )
                        value = os.path.splitext(os.path.basename(str(name)))[0]
                        return value, f"{value}.sql", lambda: False

            loader = SafeFileSystemLoader(final_search_path or [])
        else:
            loader = FileSystemLoader(final_search_path) if final_search_path else None
        extensions: list[Union[str, type[Extension]]] = ["jinja2.ext.do"]
        if self._apply_dbt_builtins(config):
            extensions.append(DBTTestExtension)

        return SandboxedEnvironment(
            # We explicitly want to preserve newlines.
            keep_trailing_newline=True,
            # The do extension allows the "do" directive
            autoescape=False,
            extensions=extensions,
            loader=loader,
        )

    def _get_macros_path(
        self, config: Optional[FluffConfig], key: str
    ) -> Optional[list[str]]:
        """Get the list of macros paths from the provided config object.

        This method searches for a config section specified by the
        templater_selector, name, and key specified. If the section is
        found, it retrieves the value associated with that section and splits it into
        a list of strings using a comma as the delimiter. The resulting list is
        stripped of whitespace and empty strings and returned. If the section is not
        found or the resulting list is empty, it returns None.

        Args:
            config (FluffConfig): The config object to search for the macros path
                section.
            key (str): Key to load the macros path from the config file.
                Also used for loading the excluding macros path from config.

        Returns:
            Optional[list[str]]: The list of macros paths if found, None otherwise.
        """
        if config:
            macros_path = config.get_section((self.templater_selector, self.name, key))
            if macros_path:
                result = [s.strip() for s in macros_path.split(",") if s.strip()]
                if result:
                    return result
        return None

    def _get_loader_search_path(
        self, config: Optional[FluffConfig]
    ) -> Optional[list[str]]:
        """Get the list of Jinja loader search paths from the provided config object.

        This method searches for a config section specified by the
        templater_selector, name, and 'loader_search_path' keys. If the section is
        found, it retrieves the value associated with that section and splits it into
        a list of strings using a comma as the delimiter. The resulting list is
        stripped of whitespace and empty strings and returned. If the section is not
        found or the resulting list is empty, it returns None.

        Args:
            config (FluffConfig): The config object to search for the loader search
                path section.

        Returns:
            Optional[list[str]]: The list of loader search paths if found, None
                otherwise.
        """
        if config:
            loader_search_path = config.get_section(
                (self.templater_selector, self.name, "loader_search_path")
            )
            if loader_search_path:
                result = [s.strip() for s in loader_search_path.split(",") if s.strip()]
                if result:
                    return result
        return None

    def _get_jinja_analyzer(self, raw_str: str, env: Environment) -> JinjaAnalyzer:
        """Creates a new object derived from JinjaAnalyzer.

        Derived classes can provide their own analyzers (e.g. to support custom Jinja
        tags).
        """
        return JinjaAnalyzer(raw_str, env)

    def _apply_dbt_builtins(self, config: Optional[FluffConfig]) -> bool:
        """Check if dbt builtins should be applied from the provided config object.

        This method searches for a config section specified by the
        templater_selector, name, and 'apply_dbt_builtins' keys. If the section
        is found, it returns the value associated with that section. If the
        section is not found, it returns False.

        Args:
            config (FluffConfig): The config object to search for the apply_dbt_builtins
                section.

        Returns:
            bool: True if dbt builtins should be applied, False otherwise.
        """
        if config:
            apply_dbt_builtins = config.get_section(
                (self.templater_selector, self.name, "apply_dbt_builtins")
            )
            # If the config is totally absent for this templater, default to False,
            # but for any other value that isn't boolean, throw an error.
            if apply_dbt_builtins is None:
                apply_dbt_builtins = False
            assert isinstance(apply_dbt_builtins, bool), (
                f"`apply_dbt_builtins` for {self.templater_selector}.{self.name} "
                f"must be True/False, not {apply_dbt_builtins!r}"
            )
            return apply_dbt_builtins
        return False

    def _get_env_context(
        self,
        fname: Optional[str],
        config: Optional[FluffConfig],
        env: Environment,
    ) -> dict[str, Any]:
        """Get the templating context from the config.

        NOTE: This closely mirrors the `get_context` method which we inherit from the
        python templater, but extends the signature. For that reason we define a new
        method here, which internally refers to `get_context`.

        Args:
            fname (str, optional): The name of the file.
            config (dict, optional): The configuration.
            env: The Jinja Environment.

        Returns:
            dict: The templating context.
        """
        # Load the context
        live_context = self.get_context(fname, config)
        # Apply dbt builtin functions if we're allowed.
        if config:
            # first make libraries available in the context
            # so they can be used by the macros too
            libraries = self._extract_libraries_from_config(config=config)
            live_context.update(libraries)

            jinja_filters = libraries.get("SQLFLUFF_JINJA_FILTERS")
            if jinja_filters:
                env.filters.update(jinja_filters)

            if self._apply_dbt_builtins(config):
                for name in DBT_BUILTINS:
                    # Only apply if it hasn't already been set at this stage.
                    if name not in live_context:
                        live_context[name] = DBT_BUILTINS[name]

        # Load macros from path (if applicable)
        if config:
            # References to variables are fixed when macros are compiled. In
            # order to handle macros that refer to other macros from another
            # file, we have to make two passes:

            # Pass 1: get all macro names and insert late-bound functions into
            # the context for every known macro.
            macro_names = self._extract_macros(
                config=config, env=env, ctx=live_context
            ).keys()
            late_binding_macros = {}
            for k in macro_names:

                def late_binding_macro(macro_name: str) -> Callable[..., Any]:
                    return lambda *args, **kwargs: live_context[macro_name](
                        *args, **kwargs
                    )

                late_binding_macros[k] = late_binding_macro(k)
            # Pass 2: load the macros, with the late bindings in the context
            live_context.update(
                self._extract_macros(
                    config=config, env=env, ctx=live_context | late_binding_macros
                )
            )

        return live_context

    def construct_render_func(
        self, fname: Optional[str] = None, config: Optional[FluffConfig] = None
    ) -> tuple[Environment, dict[str, Any], Callable[[str], str]]:
        """Builds and returns objects needed to create and run templates.

        Args:
            fname (Optional[str]): The name of the file.
            config (Optional[dict]): The configuration settings.

        Returns:
            tuple[Environment, dict, Callable[[str], str]]: A tuple
            containing the following:
                - env (Environment): An instance of the 'Environment' class.
                - live_context (dict): A dictionary containing the live context.
                - render_func (Callable[[str], str]): A callable function
                that is used to instantiate templates.
        """
        # Load the context
        env = self._get_jinja_env(config)
        live_context = self._get_env_context(fname, config, env)

        def render_func(in_str: str) -> str:
            """Used by JinjaTracer to instantiate templates.

            This function is a closure capturing internal state from process().
            Note that creating templates involves quite a bit of state known to
            _this_ function but not to JinjaTracer.

            https://www.programiz.com/python-programming/closure
            """
            try:
                template = env.from_string(in_str, globals=live_context)
            except TemplateSyntaxError as err:  # pragma: no cover
                # NOTE: If the template fails to parse, then this clause
                # will be triggered. However in normal that should never
                # happen because the template should already have been
                # validated by the point this is called. Typically that
                # happens when searching for undefined variables.
                raise SQLTemplaterError(
                    f"Late failure to parse jinja template: {err}.",
                    line_no=err.lineno,
                )
            return template.render()

        return env, live_context, render_func

    def _generate_violations_for_undefined_variables(
        self,
        in_str: str,
        syntax_tree: jinja2.nodes.Template,
        undefined_variables: set[str],
    ) -> list[SQLTemplaterError]:
        """Generates violations for any undefined variables."""
        violations: list[SQLTemplaterError] = []
        if undefined_variables:
            # Lets go through and find out where they are:
            for template_err_val in self._crawl_tree(
                syntax_tree, undefined_variables, in_str
            ):
                violations.append(template_err_val)
        return violations

    @staticmethod
    def _init_undefined_tracking(
        live_context: dict[str, Any],
        potentially_undefined_variables: Iterable[str],
        ignore_templating: bool = False,
    ) -> set[str]:
        """Sets up tracing of undefined template variables.

        NOTE: This works by mutating the `live_context` which
        is being used by the environment.
        """
        # NOTE: This set is modified by the `UndefinedRecorder` when run.
        undefined_variables: set[str] = set()

        for val in potentially_undefined_variables:
            if val not in live_context:
                if ignore_templating:
                    live_context[val] = DummyUndefined.create(val)
                else:
                    live_context[val] = UndefinedRecorder(val, undefined_variables)

        return undefined_variables

    @large_file_check
    def process(
        self,
        *,
        in_str: str,
        fname: str,
        config: Optional[FluffConfig] = None,
        formatter: Optional[FormatterInterface] = None,
    ) -> tuple[TemplatedFile, list[SQLTemplaterError]]:
        """Process a string and return the new string.

        Note that the arguments are enforced as keywords
        because Templaters can have differences in their `process`
        method signature. A Templater that only supports reading
        from a file would need the following signature:
            process(*, fname, in_str=None, config=None)
            (arguments are swapped)

        Args:
            in_str (str): The input string.
            fname (str, optional): The filename of this string. This is
                mostly for loading config files at runtime.
            config (FluffConfig): A specific config to use for this
                templating operation. Only necessary for some templaters.
            formatter (CallbackFormatter): Optional object for output.

        Raises:
            ValueError: If the 'config' argument is not provided.
            SQLTemplaterError: If templating fails fatally, then this method
                should raise a :obj:`SQLTemplaterError` instead which will be
                caught and displayed appropriately.

        Returns:
            tuple[TemplatedFile, list[SQLTemplaterError]]: A tuple containing the
            templated file and a list of violations.
        """
        if not config:  # pragma: no cover
            raise ValueError(
                "For the jinja templater, the `process()` method requires a config "
                "object."
            )

        # Fast path: a file with no Jinja markers renders to itself, so skip
        # environment construction, Jinja parsing, rendering and tracing
        # entirely. The environment always uses the standard Jinja delimiters
        # (see `_get_jinja_env`) and `keep_trailing_newline=True`, so if none
        # of `{{`, `{%` or `{#` appear then the rendered output is exactly the
        # input, there can be no undefined variables, and the file is a single
        # literal slice (matching what `slice_file` would return). We only
        # skip when no macro or library loading is configured, because
        # loading macros (from config or from a path) and loading libraries
        # (from `library_path`) is validated at construction time and can
        # raise user-facing errors which we must preserve. An empty `in_str`
        # is excluded because `TemplatedFile`'s default single-slice
        # construction produces a zero-length literal slice, which the
        # coverage check in `process_with_variants` treats as "uncovered" and
        # uses to synthesise a spurious extra variant.
        if (
            in_str
            and not re.search(r"\{[{%#]", in_str)
            and not self._get_macros_path(config, "load_macros_from_path")
            and not config.get_section((self.templater_selector, self.name, "macros"))
            and not config.get("library_path")
            and not config.get_section(
                (self.templater_selector, self.name, "library_path")
            )
        ):
            return TemplatedFile(in_str, fname=fname), []

        env, live_context, render_func = self.construct_render_func(
            fname=fname, config=config
        )

        # Attempt to identify any undeclared variables or syntax errors.
        # The majority of variables will be found during the _crawl_tree
        # step rather than this first Exception which serves only to catch
        # catastrophic errors.
        try:
            syntax_tree = env.parse(in_str)
            potentially_undefined_variables = meta.find_undeclared_variables(
                syntax_tree
            )
        except Exception as err:
            templater_error = SQLTemplaterError(
                "Failed to parse Jinja syntax. Correct the syntax or select an "
                "alternative templater. Error: " + str(err)
            )
            # Capture a line number if we can.
            if isinstance(err, TemplateSyntaxError):
                templater_error.line_no = err.lineno
            raise templater_error

        undefined_variables = self._init_undefined_tracking(
            live_context,
            potentially_undefined_variables,
            ignore_templating=("templating" in config.get("ignore")),
        )

        try:
            # Slice the file once rendered.
            raw_sliced, sliced_file, out_str = self.slice_file(
                in_str,
                render_func=render_func,
                config=config,
            )
            return (
                TemplatedFile(
                    source_str=in_str,
                    templated_str=out_str,
                    fname=fname,
                    sliced_file=sliced_file,
                    raw_sliced=raw_sliced,
                ),
                self._generate_violations_for_undefined_variables(
                    in_str, syntax_tree, undefined_variables
                ),
            )
        except (TemplateError, TypeError, ValueError) as err:
            # ValueError is caught to handle multi-variable for-loop unpacking
            # failures, e.g. {% for key, val in undefined_var.items() %} raises
            # "not enough values to unpack" because the undefined stub yields
            # only one element. We surface this as a user-friendly error rather
            # than an unhandled crash.
            templater_logger.info("Unrecoverable Jinja Error: %s", err, exc_info=True)
            raise SQLTemplaterError(
                (
                    "Unrecoverable failure in Jinja templating: {}. Have you "
                    "correctly configured your variables? "
                    "https://docs.sqlfluff.com/en/latest/perma/variables.html"
                ).format(err),
                # We don't have actual line number information, but specify
                # line 1 so users can ignore with "noqa" if they want. (The
                # default is line 0, which can't be ignored because it's not
                # a valid line number.)
                line_no=1,
                line_pos=1,
            )

    def slice_file(
        self,
        raw_str: str,
        render_func: Callable[[str], str],
        config: Optional[FluffConfig] = None,
        append_to_templated: str = "",
    ) -> tuple[list[RawFileSlice], list[TemplatedFileSlice], str]:
        """Slice the file to determine regions where we can fix.

        Args:
            raw_str (str): The raw string to be sliced.
            render_func (Callable[[str], str]): The rendering function to be used.
            config (optional): Optional configuration.
            append_to_templated: Optional string to append to the template.

        Returns:
            tuple[list[RawFileSlice], list[TemplatedFileSlice], str]:
                A tuple containing a list of raw file slices, a list of
                templated file slices, and the templated string.
        """
        # The JinjaTracer slicing algorithm is more robust, but it requires
        # us to create and render a second template (not raw_str).

        templater_logger.info("Slicing File Template")
        templater_logger.debug("    Raw String: %r", raw_str[:80])
        analyzer = self._get_jinja_analyzer(raw_str, self._get_jinja_env())
        tracer = analyzer.analyze(render_func)
        trace = tracer.trace(append_to_templated=append_to_templated)
        return trace.raw_sliced, trace.sliced_file, trace.templated_str

    @staticmethod
    def _rectify_templated_slices(
        length_deltas: dict[int, int], sliced_template: list[TemplatedFileSlice]
    ) -> list[TemplatedFileSlice]:
        """This method rectifies the source slices of a variant template.

        :TRICKY: We want to yield variants that _look like_ they were
        rendered from the original template. However, they were actually
        rendered from a modified template, which means they have source
        indices which won't line up with the source files. We correct that
        here by using the length deltas generated earlier from the
        modifications.

        This should ensure that lint issues and fixes for the variants are
        handled correctly and can be combined with those from the original
        template.
        """
        # NOTE: We sort the stack because it's important that it's in order
        # because we're going to be popping from one end of it. There's no
        # guarantee that the items are in a particular order a) because it's
        # a dict and b) because they may have been generated out of order.
        delta_stack = sorted(length_deltas.items(), key=lambda t: t[0])

        adjusted_slices: list[TemplatedFileSlice] = []
        carried_delta = 0
        for tfs in sliced_template:
            if delta_stack:
                idx, d = delta_stack[0]
                if idx == tfs.source_slice.start + carried_delta:
                    adjusted_slices.append(
                        tfs._replace(
                            # "stretch" the slice by adjusting the end more
                            # than the start.
                            source_slice=slice(
                                tfs.source_slice.start + carried_delta,
                                tfs.source_slice.stop + carried_delta - d,
                            )
                        )
                    )
                    carried_delta -= d
                    delta_stack.pop(0)
                    continue

            # No delta match. Just shift evenly.
            adjusted_slices.append(
                tfs._replace(
                    source_slice=slice(
                        tfs.source_slice.start + carried_delta,
                        tfs.source_slice.stop + carried_delta,
                    )
                )
            )
        return adjusted_slices

    @staticmethod
    def _calculate_variant_score(
        raw_sliced: list[RawFileSlice],
        sliced_file: list[TemplatedFileSlice],
        uncovered_slices: set[int],
        original_source_slices: dict[int, slice],
    ) -> int:
        """Compute a score for the variant based from size of covered slices.

        NOTE: We need to map this back to the positions in the original
        file, and only have the positions in the modified file here.
        That means we go translate back via the slice index in raw file.
        """
        # First, work out the literal positions in the modified file which
        # are now covered.
        covered_source_positions = {
            tfs.source_slice.start
            for tfs in sliced_file
            if tfs.slice_type == "literal" and not is_zero_slice(tfs.templated_slice)
        }
        # Second, convert these back into indices so we can use them to
        # refer to the unmodified source file.
        covered_raw_slice_idxs = [
            idx
            for idx, raw_slice in enumerate(raw_sliced)
            if raw_slice.source_idx in covered_source_positions
        ]

        return sum(
            slice_length(original_source_slices[idx])
            for idx in covered_raw_slice_idxs
            if idx in uncovered_slices
        )

    def _handle_unreached_code(
        self,
        in_str: str,
        render_func: Callable[[str], str],
        uncovered_slices: set[int],
        append_to_templated: str = "",
    ) -> Iterator[tuple[list[RawFileSlice], list[TemplatedFileSlice], str]]:
        """Address uncovered slices by tweaking the template to hit them.

        Args:
            in_str (:obj:`str`): The raw source file.
            render_func (:obj:`callable`): The render func for the templater.
            uncovered_slices (:obj:`set` of :obj:`int`): Indices of slices in the raw
                file which are not rendered in the original rendering. These are the
                slices we'll attempt to hit by modifying the template. NOTE: These are
                indices in the _sequence of slices_, not _character indices_ in the
                raw source file.
            append_to_templated (:obj:`str`, optional): Optional string to append
                to the templated file.
        """
        analyzer = self._get_jinja_analyzer(in_str, self._get_jinja_env())
        tracer_copy = analyzer.analyze(render_func)

        max_variants_generated = 10
        max_variants_returned = 5
        variants: dict[str, tuple[int, JinjaTrace, dict[int, int]]] = {}

        # Create a mapping of the original source slices before modification so
        # we can adjust the positions post-modification.
        original_source_slices = {
            idx: raw_slice.source_slice()
            for idx, raw_slice in enumerate(tracer_copy.raw_sliced)
        }

        for uncovered_slice in sorted(uncovered_slices)[:max_variants_generated]:
            tracer_probe = copy.deepcopy(tracer_copy)
            tracer_trace = copy.deepcopy(tracer_copy)
            override_raw_slices = []
            # `length_deltas` is to keep track of the length changes associated
            # with the changes we're making so we can correct the positions in
            # the resulting template.
            length_deltas: dict[int, int] = {}
            # Find a path that takes us to 'uncovered_slice'.
            choices = tracer_probe.move_to_slice(uncovered_slice, 0)
            for branch, options in choices.items():
                raw_file_slice = tracer_probe.raw_sliced[branch]
                if raw_file_slice.tag in ("if", "elif"):
                    # Replace the existing "if" of "elif" expression with a new,
                    # hardcoded value that hits the target slice in the template
                    # (here that is options[0]).
                    new_value = "True" if options[0] == branch + 1 else "False"
                    new_source = f"{{% {raw_file_slice.tag} {new_value} %}}"
                    tracer_trace.raw_slice_info[
                        raw_file_slice
                    ].alternate_code = new_source
                    override_raw_slices.append(branch)
                    length_deltas[raw_file_slice.source_idx] = len(new_source) - len(
                        raw_file_slice.raw
                    )

            # Render and analyze the template with the overrides.
            variant_key = tuple(
                (
                    cast(str, tracer_trace.raw_slice_info[rs].alternate_code)
                    if idx in override_raw_slices
                    and tracer_trace.raw_slice_info[rs].alternate_code is not None
                    else rs.raw
                )
                for idx, rs in enumerate(tracer_trace.raw_sliced)
            )
            # In some cases (especially with nested if statements), we may
            # generate a variant that duplicates an existing variant. Skip
            # those.
            variant_raw_str = "".join(variant_key)
            if variant_raw_str not in variants:
                analyzer = self._get_jinja_analyzer(
                    variant_raw_str, self._get_jinja_env()
                )
                tracer_trace = analyzer.analyze(render_func)
                try:
                    trace = tracer_trace.trace(
                        append_to_templated=append_to_templated,
                    )
                except Exception:
                    # If we get an error tracing the variant, skip it. This may
                    # happen for a variety of reasons. Basically there's no
                    # guarantee that the variant will be valid Jinja.
                    continue
                else:
                    # Compute a score for the variant based on the size of initially
                    # uncovered literal slices it hits.
                    score = self._calculate_variant_score(
                        raw_sliced=trace.raw_sliced,
                        sliced_file=trace.sliced_file,
                        uncovered_slices=uncovered_slices,
                        original_source_slices=original_source_slices,
                    )

                    variants[variant_raw_str] = (score, trace, length_deltas)

        # Return the top-scoring variants.
        sorted_variants: list[tuple[int, JinjaTrace, dict[int, int]]] = sorted(
            variants.values(), key=lambda v: v[0], reverse=True
        )
        for _, trace, deltas in sorted_variants[:max_variants_returned]:
            # Rectify the source slices of the generated template, which should
            # ensure that lint issues and fixes for the variants are handled
            # correctly and can be combined with those from the original template.
            adjusted_slices = self._rectify_templated_slices(
                deltas,
                trace.sliced_file,
            )
            yield (
                tracer_copy.raw_sliced,
                adjusted_slices,
                trace.templated_str,
            )

    @large_file_check
    def process_with_variants(
        self,
        *,
        in_str: str,
        fname: str,
        config: Optional[FluffConfig] = None,
        formatter: Optional[FormatterInterface] = None,
    ) -> Iterator[tuple[TemplatedFile, list[SQLTemplaterError]]]:
        """Process a string and return one or more variant renderings.

        Note that the arguments are enforced as keywords
        because Templaters can have differences in their
        `process` method signature.
        A Templater that only supports reading from a file
        would need the following signature:
            process(*, fname, in_str=None, config=None)
        (arguments are swapped)

        Args:
            in_str (:obj:`str`): The input string.
            fname (:obj:`str`, optional): The filename of this string. This is
                mostly for loading config files at runtime.
            config (:obj:`FluffConfig`): A specific config to use for this
                templating operation. Only necessary for some templaters.
            formatter (:obj:`CallbackFormatter`): Optional object for output.

        """
        templated_file, violations = self.process(
            in_str=in_str, fname=fname, config=config, formatter=formatter
        )
        yield templated_file, violations

        # Find uncovered code (if any), tweak the template to hit that code.
        # First, identify the literals which _are_ covered.
        covered_literal_positions = {
            tfs.source_slice.start
            for tfs in templated_file.sliced_file
            # It's covered if it's rendered
            if not is_zero_slice(tfs.templated_slice)
        }
        templater_logger.debug(
            "Covered literal positions %s", covered_literal_positions
        )

        uncovered_literal_idxs = {
            idx
            for idx, raw_slice in enumerate(templated_file.raw_sliced)
            if raw_slice.slice_type == "literal"
            and raw_slice.source_idx not in covered_literal_positions
            and not self._is_unreached_whitespace_adjacent_to_trim_tag(
                templated_file.raw_sliced, idx
            )
        }
        templater_logger.debug(
            "Uncovered literals correspond to slices %s", uncovered_literal_idxs
        )

        # If every literal is already covered, there's no unreached code to
        # find, so there are no variants beyond the one already yielded above.
        # Skip building a render function entirely in that case (e.g. a
        # literal-only file handled by the fast path in `process()`).
        if not uncovered_literal_idxs:
            return

        # NOTE: No validation required as all validation done in the `.process()`
        # call above.
        _, _, render_func = self.construct_render_func(fname=fname, config=config)

        for raw_sliced, sliced_file, templated_str in self._handle_unreached_code(
            in_str, render_func, uncovered_literal_idxs
        ):
            yield (
                TemplatedFile(
                    source_str=in_str,
                    templated_str=templated_str,
                    fname=fname,
                    sliced_file=sliced_file,
                    raw_sliced=raw_sliced,
                ),
                violations,
            )

    @staticmethod
    def _exclude_macros(macro_path: str, exclude_macros_path: list[str]) -> bool:
        """Determines if a macro is within the exclude macros path.

        These macros will be ignored and not loaded into context

        Args:
            macro_path (str): Str of the path to the macro
            exclude_macros_path (list[str]): Str of the path to the macros to exclude

        Returns:
            bool: True if the macro should be excluded
        """
        for exclude_path in exclude_macros_path:
            macro_path_normalized = os.path.normpath(os.path.abspath(macro_path))
            exclude_path_normalized = os.path.normpath(exclude_path)
            if exclude_path_normalized in macro_path_normalized:
                templater_logger.debug("Skipping this macro file: %s", macro_path)
                return True
        return False


class DummyUndefined(jinja2.Undefined):
    """Acts as a dummy value to try and avoid template failures.

    Inherits from jinja2.Undefined so Jinja's default() filter will
    treat it as a missing value, even though it has a non-empty value
    in normal contexts.
    """

    # Tell Jinja this object is safe to call and does not alter data.
    # https://jinja.palletsprojects.com/en/2.9.x/sandbox/#jinja2.sandbox.SandboxedEnvironment.is_safe_callable
    unsafe_callable = False
    # https://jinja.palletsprojects.com/en/3.0.x/sandbox/#jinja2.sandbox.SandboxedEnvironment.is_safe_callable
    alters_data = False

    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

    def __str__(self) -> str:
        return self.name.replace(".", "_")

    @classmethod
    def create(cls, name: str) -> "DummyUndefined":
        """Factory method.

        When ignoring=templating is configured, use 'name' as the value for
        undefined variables. We deliberately avoid recording and reporting
        undefined variables as errors. Using 'name' as the value won't always
        work, but using 'name', combined with implementing the magic methods
        (such as __eq__, see above), works well in most cases.
        """
        templater_logger.debug(
            "Providing dummy value for undefined Jinja variable: %s", name
        )
        result = DummyUndefined(name)
        return result

    def __getattr__(self, item: str) -> "DummyUndefined":
        """Intercept any calls to undefined attributes.

        Args:
            item (str): The name of the attribute.

        Returns:
            object: A dynamically created instance of this class.
        """
        return self.create(f"{self.name}.{item}")

    # Implement the most common magic methods. This helps avoid
    # templating errors for undefined variables.
    # https://www.tutorialsteacher.com/python/magic-methods-in-python
    def _self_impl(self, *args: Any, **kwargs: Any) -> "DummyUndefined":
        """Return an instance of the class itself.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            object: An instance of the class itself.
        """
        return self

    def _bool_impl(self, *args: Any, **kwargs: Any) -> bool:
        """Return a boolean value.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            bool: A boolean value.
        """
        return True

    # We're intentionally changing the behaviour here compared to the base
    # class of Undefined. That means we're going to silence the `assignment`
    # mypy warnings. Operations on an undefined result in another undefined.
    __add__ = __sub__ = __mul__ = _self_impl  # type: ignore[assignment]
    __floordiv__ = __truediv__ = _self_impl  # type: ignore[assignment]
    __mod__ = __pow__ = _self_impl  # type: ignore[assignment]
    __pos__ = __neg__ = _self_impl  # type: ignore[assignment]
    __lshift__ = __rshift__ = _self_impl
    __getitem__ = _self_impl  # type: ignore[assignment]
    __invert__ = _self_impl
    __call__ = _self_impl  # type: ignore[assignment]
    # Boolean operations on an undefined are handled separately.
    __and__ = __or__ = __xor__ = __bool__ = _bool_impl
    __lt__ = __le__ = __ge__ = __gt__ = _bool_impl  # type: ignore[assignment]
    __eq__ = __ne__ = _bool_impl

    def __hash__(self) -> int:  # pragma: no cover
        """Return a constant hash value.

        Returns:
            int: A constant hash value.
        """
        # This is called by the "in" operator, among other things.
        return 0

    def __iter__(self) -> Iterator["DummyUndefined"]:
        """Return an iterator that contains only the instance of the class itself.

        Returns:
            iterator: An iterator.
        """
        return [self].__iter__()


class DBTTestExtension(Extension):
    """Jinja extension to handle the dbt test tag."""

    tags = {"test"}

    def parse(self, parser: jinja2.parser.Parser) -> jinja2.nodes.Macro:
        """Parses out the contents of the test tag."""
        node = jinja2.nodes.Macro(lineno=next(parser.stream).lineno)
        test_name = parser.parse_assign_target(name_only=True).name

        parser.parse_signature(node)
        node.name = f"test_{test_name}"
        node.body = parser.parse_statements(("name:endtest",), drop_needle=True)
        return node



"""Source for the MatchResult class.

This should be the default response from any `match` method.
"""

from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, DefaultDict, Optional, Union

from sqlfluff.core.helpers.slice import slice_length
from sqlfluff.core.parser.markers import PositionMarker

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.parser.context import ParseContext
    from sqlfluff.core.parser.segments import BaseSegment, MetaSegment


def _get_point_pos_at_idx(
    segments: Sequence["BaseSegment"], idx: int
) -> PositionMarker:
    if idx < len(segments):
        _next_pos = segments[idx].pos_marker
        assert _next_pos, "Segments passed to .apply() should all have position."
        return _next_pos.start_point_marker()
    else:
        _prev_pos = segments[idx - 1].pos_marker
        assert _prev_pos, "Segments passed to .apply() should all have position."
        return _prev_pos.end_point_marker()


@dataclass(frozen=True)
class MatchResult:
    """This should be the default response from any `match` method.

    All references and indices are in reference to a single root tuple
    of segments. This result contains enough information to actually
    create the nested tree structure, but shouldn't actually contain
    any new segments itself. That means keeping information about:
    1. Ranges of segments which should be included segments to be
       created.
    2. References to the segment classes which we would create.
    3. Information about any _new_ segments to add in the process,
       such as MetaSegment classes.

    Given the segments aren't yet "nested", the structure of this
    result *will* need to be nested, ideally self nested.

    In the case of finding unparsable locations, we should return the
    "best" result, referencing the furthest that we got. That allows
    us to identify those parsing issues and create UnparsableSegment
    classes later.
    """

    # Slice in the reference tuple
    matched_slice: slice
    # Reference to the kind of segment to create.
    # NOTE: If this is null, it means we've matched a sequence of segments
    # but not yet created a container to put them in.
    matched_class: Optional[type["BaseSegment"]] = None
    # kwargs to pass to the segment on creation.
    segment_kwargs: dict[str, Any] = field(default_factory=dict)
    # Types and indices to add in new segments (they'll be meta segments)
    insert_segments: tuple[tuple[int, type["MetaSegment"]], ...] = field(
        default_factory=tuple
    )
    # Child segment matches (this is the recursive bit)
    child_matches: tuple["MatchResult", ...] = field(default_factory=tuple)

    def __post_init__(self) -> None:
        """Do some lightweight validation post instantiation."""
        if not slice_length(self.matched_slice):
            # Zero length matches with inserts are allowed, but not with
            # matched_class or child_matches.
            assert not self.matched_class, (
                "Tried to create zero length MatchResult with "
                "`matched_class`. This MatchResult is invalid. "
                f"{self.matched_class} @{self.matched_slice}"
            )
            assert not self.child_matches, (
                "Tried to create zero length MatchResult with "
                "`child_matches`. Is this allowed?! "
                f"Result: {self}"
            )

    def __len__(self) -> int:
        return slice_length(self.matched_slice)

    def __bool__(self) -> bool:
        """A MatchResult is truthy if it has length or inserts."""
        return len(self) > 0 or bool(self.insert_segments)

    def __str__(self) -> str:
        return self.stringify()

    def stringify(self, indent: str = "") -> str:
        """Pretty print a match for debugging."""
        prefix = f"Match ({self.matched_class}): {self.matched_slice}"
        buffer = prefix
        for key, value in self.segment_kwargs.items():
            buffer += f"\n  {indent}-{key}: {value!r}"
        if self.insert_segments:
            for idx, insert in self.insert_segments:
                buffer += f"\n  {indent}+{idx}: {insert}"
        if self.child_matches:
            for child in self.child_matches:
                buffer += f"\n  {indent}+{child.stringify(indent + '  ')}"
        return buffer

    @classmethod
    def empty_at(cls, idx: int) -> "MatchResult":
        """Create an empty match at a particular index."""
        return cls(slice(idx, idx))

    def is_better_than(self, other: "MatchResult") -> bool:
        """A match is better compared on length."""
        return len(self) > len(other)

    def append(
        self,
        other: "MatchResult",
        insert_segments: tuple[tuple[int, type["MetaSegment"]], ...] = (),
    ) -> "MatchResult":
        """Combine another subsequent match onto this one.

        NOTE: Because MatchResult is frozen, this returns a new
        match.
        """
        # If the current match is empty, just return the other.
        if not len(self) and not self.insert_segments:
            return other
        # If the same is true of the other, just return self.
        if not len(other) and not other.insert_segments:
            return self  # pragma: no cover

        # Otherwise the two must follow each other.
        # NOTE: A gap is allowed, but is assumed to be included in the
        # match.
        assert self.matched_slice.stop <= other.matched_slice.start
        new_slice = slice(self.matched_slice.start, other.matched_slice.stop)
        child_matches: tuple[MatchResult, ...] = ()
        for match in (self, other):
            # If it's got a matched class, add it as a child.
            if match.matched_class:
                child_matches += (match,)
            # Otherwise incorporate
            else:
                # Note: We're appending to the optional insert segments
                # provided in the kwargs.
                insert_segments += match.insert_segments
                child_matches += match.child_matches
        return MatchResult(
            new_slice,
            insert_segments=insert_segments,
            child_matches=child_matches,
        )

    def wrap(
        self,
        outer_class: type["BaseSegment"],
        insert_segments: tuple[tuple[int, type["MetaSegment"]], ...] = (),
        segment_kwargs: dict[str, Any] = {},
    ) -> "MatchResult":
        """Wrap this result with an outer class.

        NOTE: Because MatchResult is frozen, this returns a new
        match.
        """
        # If it's a failed (empty) match, then just pass straight
        # through. It's not valid to add a matched class to an empty
        # result.
        if not slice_length(self.matched_slice) and not self.insert_segments:
            assert not insert_segments, "Cannot wrap inserts onto an empty match."
            return self

        child_matches: tuple[MatchResult, ...]
        if self.matched_class:
            # If the match already has a class, then make
            # the current one and child match and clear the
            # other buffers.
            child_matches = (self,)
        else:
            # Otherwise flatten the existing match into
            # the new one.
            insert_segments = self.insert_segments + insert_segments
            child_matches = self.child_matches

        # Otherwise flatten the content
        return MatchResult(
            self.matched_slice,
            matched_class=outer_class,
            segment_kwargs=segment_kwargs,
            insert_segments=insert_segments,
            child_matches=child_matches,
        )

    def apply(
        self,
        segments: tuple["BaseSegment", ...],
        parse_context: Optional["ParseContext"] = None,
    ) -> tuple["BaseSegment", ...]:
        """Actually this match to segments to instantiate.

        This turns a theoretical match into a nested structure of segments.

        We handle child segments _first_ so that we can then include them when
        creating the parent. That means sequentially working through the children
        and any inserts. If there are overlaps, then we have a problem, and we
        should abort.
        """
        # NOTE: Accumulate into a list and build the tuple once at the end.
        # `tuple += ...` looks equivalent but isn't: tuples are immutable, so
        # each `+=` allocates a new tuple and copies everything accumulated
        # so far, making a loop of N appends O(N^2). A list amortises
        # appends/extends to O(N), which matters here since this runs at
        # every level of a potentially large tree (e.g. long column/UNION/IN
        # lists produce nodes with many trigger locations).
        result_segments_list: list["BaseSegment"] = []
        if not slice_length(self.matched_slice):
            assert not self.matched_class, (
                "Tried to apply zero length MatchResult with "
                "`matched_class`. This MatchResult is invalid. "
                f"{self.matched_class} @{self.matched_slice}"
            )
            assert not self.child_matches, (
                "Tried to apply zero length MatchResult with "
                "`child_matches`. This MatchResult is invalid. "
                f"Result: {self}"
            )
            if self.insert_segments:
                assert segments, "Cannot insert segments without reference position."
                for idx, seg in self.insert_segments:
                    assert idx == self.matched_slice.start, (
                        f"Tried to insert @{idx} outside of matched "
                        f"slice {self.matched_slice}"
                    )
                    _pos = _get_point_pos_at_idx(segments, idx)
                    if parse_context:
                        parse_context.increment_parse_nodes()
                    result_segments_list.append(seg(pos_marker=_pos))
            return tuple(result_segments_list)

        assert len(segments) >= self.matched_slice.stop, (
            f"Matched slice ({self.matched_slice}) sits outside segment "
            f"bounds: {len(segments)}"
        )

        # Which are the locations we need to care about?
        trigger_locs: DefaultDict[
            int, list[Union[MatchResult, type["MetaSegment"]]]
        ] = defaultdict(list)
        # Add the inserts first...
        for insert in self.insert_segments:
            trigger_locs[insert[0]].append(insert[1])
        # ...and then the matches
        for match in self.child_matches:
            trigger_locs[match.matched_slice.start].append(match)

        # Then work through creating any subsegments.
        max_idx = self.matched_slice.start
        for idx in sorted(trigger_locs.keys()):
            # Have we passed any untouched segments?
            if idx > max_idx:
                # If so, add them in unchanged.
                result_segments_list.extend(segments[max_idx:idx])
                max_idx = idx
            elif idx < max_idx:  # pragma: no cover
                raise ValueError(
                    "Segment skip ahead error. An outer match contains "
                    "overlapping child matches. This MatchResult was "
                    "wrongly constructed."
                )
            # Then work through each of the triggers.
            for trigger in trigger_locs[idx]:
                # If it's a match, apply it.
                if isinstance(trigger, MatchResult):
                    result_segments_list.extend(
                        trigger.apply(segments=segments, parse_context=parse_context)
                    )
                    # Update the end slice.
                    max_idx = trigger.matched_slice.stop
                    continue

                # Otherwise it's a segment.
                # Get the location from the next segment unless there isn't one.
                _pos = _get_point_pos_at_idx(segments, idx)
                result_segments_list.append(trigger(pos_marker=_pos))

        # If we finish working through the triggers and there's
        # still something left, then add that too.
        if max_idx < self.matched_slice.stop:
            result_segments_list.extend(segments[max_idx : self.matched_slice.stop])

        result_segments = tuple(result_segments_list)

        if not self.matched_class:
            return result_segments

        # Otherwise construct the subsegment
        new_seg: "BaseSegment" = self.matched_class.from_result_segments(
            result_segments, self.segment_kwargs
        )
        if parse_context:
            parse_context.increment_parse_nodes()
        return (new_seg,)



"""Methods for loading config from pyproject.toml files."""

import re
import sys
from typing import Any, TypeVar

if sys.version_info >= (3, 11):
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib

from sqlfluff.core.errors import SQLFluffUserError
from sqlfluff.core.helpers.dict import (
    NestedDictRecord,
    iter_records_from_nested_dict,
    records_to_nested_dict,
)
from sqlfluff.core.types import ConfigMappingType

T = TypeVar("T")


def _condense_rule_record(record: NestedDictRecord[T]) -> NestedDictRecord[T]:
    """Helper function to condense the rule section of a toml config."""
    key, value = record
    if len(key) > 2:
        key = (".".join(key[:-1]), key[-1])
    return key, value


def _validate_structure(raw_config: dict[str, Any]) -> ConfigMappingType:
    """Helper function to narrow types for use by SQLFluff.

    This is a recursive function on any dict keys found.
    """
    validated_config: ConfigMappingType = {}
    for key, value in raw_config.items():
        if isinstance(value, dict):
            validated_config[key] = _validate_structure(value)
        elif isinstance(value, list):
            # Coerce all list items to strings, to be in line
            # with the behaviour of ini configs.
            validated_config[key] = [str(item) for item in value]
        elif isinstance(value, (str, int, float, bool)) or value is None:
            validated_config[key] = value
        else:  # pragma: no cover
            # Whatever we found, make it into a string.
            # This is very unlikely to happen and is more for completeness.
            validated_config[key] = str(value)
    return validated_config


def _format_toml_parse_error(
    filepath: str, raw_toml: str, err: tomllib.TOMLDecodeError
) -> str:
    """Build a user-facing TOML parsing error message."""
    raw_message = str(err)
    message = getattr(err, "msg", raw_message)
    line = getattr(err, "lineno", None)
    column = getattr(err, "colno", None)

    if not hasattr(err, "msg"):
        location_match = re.search(
            r" \(at line (?P<line>\d+), column (?P<column>\d+)\)$", raw_message
        )
        if location_match:
            message = raw_message[: location_match.start()]
            line = line or int(location_match.group("line"))
            column = column or int(location_match.group("column"))

    location = ""
    if line is not None and column is not None:
        location = f" at line {line}, column {column}"
    elif line is not None:
        location = f" at line {line}"

    hint = ""
    if raw_toml.startswith("\ufeff"):
        hint = " pyproject.toml contains a UTF-8 BOM; save it as UTF-8 without BOM."

    return f"Failed to parse TOML config file {filepath}: {message}{location}.{hint}"


def load_toml_file_config(filepath: str) -> ConfigMappingType:
    """Read the SQLFluff config section of a pyproject.toml file.

    We don't need to change any key names here, because the root
    section of the toml file format is `tool.sqlfluff.core`.

    NOTE: Toml files are always encoded in UTF-8. That is a necessary
    part of the toml spec: https://toml.io/en/v1.0.0
    """
    with open(filepath, mode="r", encoding="utf-8") as file:
        raw_toml = file.read()
    try:
        toml_dict = tomllib.loads(raw_toml)
    except tomllib.TOMLDecodeError as err:
        raise SQLFluffUserError(
            _format_toml_parse_error(filepath, raw_toml, err)
        ) from None
    config_dict = _validate_structure(toml_dict.get("tool", {}).get("sqlfluff", {}))

    # NOTE: For the "rules" section of the sqlfluff config,
    # rule names are often qualified with a dot ".". In the
    # toml scenario this can get interpreted as a nested
    # section, and we resolve that edge case here.
    if "rules" not in config_dict:
        # No rules section, so no need to resolve.
        return config_dict

    rules_section = config_dict["rules"]
    assert isinstance(rules_section, dict), (
        "Expected to find section in `rules` section of config, "
        f"but instead found {rules_section}"
    )
    # Condense the rules section.
    config_dict["rules"] = records_to_nested_dict(
        _condense_rule_record(record)
        for record in iter_records_from_nested_dict(rules_section)
    )

    return config_dict



"""The parser context.

This mirrors some of the same design of the flask
context manager. https://flask.palletsprojects.com/en/1.1.x/

The context acts as a way of keeping track of state, references
to common configuration and dialects, logging and also the parse
and match depth of the current operation.
"""

import logging
import uuid
from collections import defaultdict
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, NoReturn, Optional

from tqdm import tqdm

from sqlfluff.core.config import progress_bar_configuration
from sqlfluff.core.errors import SQLParseError

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.config import FluffConfig
    from sqlfluff.core.dialects.base import Dialect
    from sqlfluff.core.parser.match_result import MatchResult
    from sqlfluff.core.parser.matchable import Matchable

# Get the parser logger
parser_logger = logging.getLogger("sqlfluff.parser")


class ParseContext:
    """Object to handle the context at hand during parsing.

    Holds two tiers of references.
    1. Persistent config, like references to the dialect or
       the current verbosity and logger.
    2. Stack config, like the parse and match depth.

    The manipulation of the stack config is done using a context
    manager and layered config objects inside the context.

    NOTE: We use context managers here to avoid _copying_
    the context, just to mutate it safely. This is significantly
    more performant than the copy operation, but does require some
    care to use properly.

    When fetching elements from the context, we first look
    at the top level stack config object and the persistent
    config values (stored as attributes of the ParseContext
    itself).
    """

    def __init__(
        self,
        dialect: "Dialect",
        max_parse_depth: int,
        max_parse_nodes: int = 0,
        indentation_config: Optional[dict[str, Any]] = None,
    ) -> None:
        """Initialize a new instance of the class.

        Args:
            dialect (Dialect): The dialect used for parsing.
            max_parse_depth (int): Maximum match depth; ``0`` to disable.
            max_parse_nodes (int): Maximum number of parse tree nodes; ``0`` to
                disable.
            indentation_config (Optional[dict[str, Any]], optional): The indentation
                configuration used by Indent and Dedent to control the intended
                indentation of certain features. Defaults to None.
        """
        self.dialect = dialect
        # Indentation config is used by Indent and Dedent and used to control
        # the intended indentation of certain features. Specifically it is
        # used in the Conditional grammar.
        self.indentation_config = indentation_config or {}
        self.max_parse_depth = max_parse_depth
        self.max_parse_nodes = max_parse_nodes
        self.current_parse_nodes = 0
        # This is the logger that child objects will latch onto.
        self.logger = parser_logger
        # A uuid for this parse context to enable cache invalidation
        self.uuid = uuid.uuid4()
        # A dict for parse caching. This is reset for each file,
        # but persists for the duration of an individual file parse.
        self._parse_cache: dict[tuple[Any, ...], "MatchResult"] = {}
        # A dictionary for keeping track of some statistics on parsing
        # for performance optimisation.
        # Focused around BaseGrammar._longest_trimmed_match().
        # Initialise only with "next_counts", the rest will be int
        # and are dealt with in .increment().
        self.parse_stats: dict[str, Any] = {"next_counts": defaultdict(int)}
        # The following attributes are only accessible via a copy
        # and not in the init method.
        # NOTE: We default to the name `File` which is not
        # particularly informative, does indicate the root segment.
        self.match_segment: str = "File"
        self._match_stack: list[str] = []
        self._parse_stack: list[str] = []
        self.match_depth = 0
        self.parse_depth = 0
        # self.terminators is a tuple to afford some level of isolation
        # and protection from edits to outside the context. This introduces
        # a little more overhead than a list, but we manage this by only
        # copying it when necessary.
        # NOTE: Includes inherited parent terminators.
        self.terminators: tuple["Matchable", ...] = ()
        # Value for holding a reference to the progress bar.
        self._tqdm: Optional[tqdm[NoReturn]] = None
        # Variable to store whether we're tracking progress. When looking
        # ahead to terminators or suchlike, we set this to False so as not
        # to confuse the progress bar.
        self.track_progress = True
        # The current character, to store where the progress bar is at.
        self._current_char = 0

    @classmethod
    def from_config(cls, config: "FluffConfig") -> "ParseContext":
        """Construct a `ParseContext` from a `FluffConfig`.

        Args:
            config (FluffConfig): The configuration object.

        Returns:
            ParseContext: The constructed ParseContext object.
        """
        indentation_config = config.get_section("indentation") or {}
        try:
            indentation_config = {k: bool(v) for k, v in indentation_config.items()}
        except TypeError:  # pragma: no cover
            raise TypeError(
                "One of the configuration keys in the `indentation` section is not "
                "True or False: {!r}".format(indentation_config)
            )
        max_parse_depth = config.get("max_parse_depth")
        max_parse_nodes = config.get("max_parse_nodes")
        assert isinstance(max_parse_depth, int)
        assert isinstance(max_parse_nodes, int)
        return cls(
            dialect=config.get("dialect_obj"),
            indentation_config=indentation_config,
            max_parse_depth=max_parse_depth,
            max_parse_nodes=max_parse_nodes,
        )

    def increment_parse_nodes(self, count: int = 1) -> None:
        """Increment the number of materialized parse nodes.

        Args:
            count (int): Number of nodes to add to the current budget.
        """
        assert count >= 0
        self.current_parse_nodes += count
        if self.max_parse_nodes > 0 and self.current_parse_nodes > self.max_parse_nodes:
            raise SQLParseError(
                f"Maximum parse node count exceeded (limit {self.max_parse_nodes}). "
                "This may indicate unusually large SQL or a malicious input."
            )

    def seed_parse_nodes(self, count: int) -> None:
        """Seed the current node budget from an existing segment count."""
        self.increment_parse_nodes(count)

    def _set_terminators(
        self,
        clear_terminators: bool = False,
        push_terminators: Optional[Sequence["Matchable"]] = None,
    ) -> tuple[int, tuple["Matchable", ...]]:
        """Set the terminators used in the class.

        This private method sets the terminators used in the class. If
        `clear_terminators` is True and the existing terminators are not
        already cleared, the method clears the terminators. If `push_terminators` is
        provided, the method appends them to the existing terminators if they are not
        already present.

        Args:
            clear_terminators (bool, optional): A flag indicating whether to clear the
                existing terminators. Defaults to False.
            push_terminators (Optional[Sequence["Matchable"]], optional): A sequence of
                `Matchable` objects to be added as terminators.
            Defaults to None.

        Returns:
            tuple[int, tuple["Matchable", ...]]: A tuple containing the
            number of terminators appended and the original terminators.
        """
        _appended = 0
        # Retain a reference to the original terminators.
        _terminators = self.terminators
        # Note: only need to reset if clear _and not already clear_.
        if clear_terminators and self.terminators:
            # NOTE: It's really important that we .copy() on the way in, because
            # we don't know what else has a reference to the input list, and
            # we rely a lot in this code on having full control over the
            # list of terminators.
            self.terminators = tuple(push_terminators) if push_terminators else ()
        elif push_terminators:
            # Yes, inefficient for now.
            for terminator in push_terminators:
                if terminator not in self.terminators:
                    self.terminators += (terminator,)
                    _appended += 1
        return _appended, _terminators

    def _reset_terminators(
        self,
        appended: int,
        terminators: tuple["Matchable", ...],
        clear_terminators: bool = False,
    ) -> None:
        """Reset the terminators attribute of the class.

        This method is used to reset the terminators attribute of the
        class. If the clear_terminators parameter is True, the terminators attribute
        is set to the provided terminators. If the clear_terminators parameter is
        False and the appended parameter is non-zero, the terminators attribute is
        trimmed to its original length minus the value of the appended parameter.

        Args:
            appended (int): The number of terminators that were appended.
            terminators (tuple["Matchable", ...]): The original terminators.
            clear_terminators (bool, optional): If True, clear the terminators attribute
                completely. Defaults to False.
        """
        # If we totally reset them, just reinstate the old object.
        if clear_terminators:
            self.terminators = terminators
        # If we didn't, then trim any added ones.
        # NOTE: Because we dedupe, just because we had push_terminators
        # doesn't mean any of them actually got added here - we only trim
        # the number that actually got appended.
        elif appended:
            # Trim back to original length.
            self.terminators = self.terminators[:-appended]

    @contextmanager
    def deeper_match(
        self,
        name: str,
        clear_terminators: bool = False,
        push_terminators: Optional[Sequence["Matchable"]] = None,
        track_progress: Optional[bool] = None,
    ) -> Iterator["ParseContext"]:
        """Increment match depth.

        Args:
            name (:obj:`str`): Name of segment we are starting to parse.
                NOTE: This value is entirely used for tracking and logging
                purposes.
            clear_terminators (:obj:`bool`, optional): Whether to force
                clear any inherited terminators. This is useful in structures
                like brackets, where outer terminators shouldn't apply while
                within. Terminators are stashed until we return back out of
                this context.
            push_terminators (:obj:`Sequence` of :obj:`Matchable`): Additional
                terminators to add to the environment while in this context.
            track_progress (:obj:`bool`, optional): Whether to pause progress
                tracking for deeper matches. This avoids having the linting
                progress bar jump forward when performing greedy matches on
                terminators.
        """
        self._match_stack.append(self.match_segment)
        self.match_segment = name
        self.match_depth += 1
        if self.max_parse_depth > 0 and self.match_depth > self.max_parse_depth:
            raise SQLParseError(
                f"Maximum parse depth exceeded (limit {self.max_parse_depth}). "
                "This may indicate deeply nested SQL or a malicious input."
            )
        _append, _terms = self._set_terminators(clear_terminators, push_terminators)
        _track_progress = self.track_progress
        if track_progress is False:
            self.track_progress = False
        elif track_progress is True:  # pragma: no cover
            # We can't go from False to True. Raise an issue if not.
            assert self.track_progress is True, "Cannot set tracking from False to True"
        try:
            yield self
        finally:
            self._reset_terminators(
                _append, _terms, clear_terminators=clear_terminators
            )
            self.match_depth -= 1
            # Reset back to old name
            self.match_segment = self._match_stack.pop()
            # Reset back to old progress tracking.
            self.track_progress = _track_progress

    @contextmanager
    def progress_bar(self, last_char: int) -> Iterator["ParseContext"]:
        """Set up the progress bar (if it's not already set up).

        Args:
            last_char (:obj:`int`): The templated character position of the
                final segment in the sequence. This is usually populated
                from the end of `templated_slice` on the final segment.
                We require this on initialising the progress bar so that
                we know how far there is to go as we track progress through
                the file.
        """
        assert not self._tqdm, "Attempted to re-initialise progressbar."
        self._tqdm = tqdm(
            # Progress is character by character in the *templated* file.
            total=last_char,
            desc="parsing",
            miniters=1,
            mininterval=0.2,
            disable=progress_bar_configuration.disable_progress_bar,
            leave=False,
        )

        try:
            yield self
        finally:
            self._tqdm.close()

    def update_progress(self, char_idx: int) -> None:
        """Update the progress bar if configured.

        If progress isn't configured, we do nothing.
        If `track_progress` is false we do nothing.
        """
        if not self._tqdm or not self.track_progress:
            return None
        if char_idx <= self._current_char:
            return None
        self._tqdm.update(char_idx - self._current_char)
        self._current_char = char_idx
        return None

    def stack(self) -> tuple[tuple[str, ...], tuple[str, ...]]:  # pragma: no cover
        """Return stacks as a tuples so that it can't be edited."""
        return tuple(self._parse_stack), tuple(self._match_stack)

    def check_parse_cache(
        self, loc_key: tuple[Any, ...], matcher_key: str
    ) -> Optional["MatchResult"]:
        """Check against the parse cache for a pre-existing match.

        If no match is found in the cache, this returns None.
        """
        return self._parse_cache.get((loc_key, matcher_key))

    def put_parse_cache(
        self, loc_key: tuple[Any, ...], matcher_key: str, match: "MatchResult"
    ) -> None:
        """Store a match in the cache for later retrieval."""
        self._parse_cache[(loc_key, matcher_key)] = match



"""The code for the Lexer."""

import logging
from collections.abc import Iterator
from typing import Any, NamedTuple, Optional, Union
from uuid import UUID, uuid4

import regex

from sqlfluff.core.config import FluffConfig
from sqlfluff.core.errors import SQLLexError
from sqlfluff.core.helpers.slice import is_zero_slice, offset_slice, to_tuple
from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.core.parser.segments import (
    BaseSegment,
    Dedent,
    EndOfFile,
    ImplicitIndent,
    Indent,
    LiteralKeywordSegment,
    MetaSegment,
    RawSegment,
    TemplateLoop,
    TemplateSegment,
    UnlexableSegment,
)
from sqlfluff.core.templaters import TemplatedFile
from sqlfluff.core.templaters.base import TemplatedFileSlice

# Instantiate the lexer logger
lexer_logger = logging.getLogger("sqlfluff.lexer")


class BlockTracker:
    """This is an object for keeping track of templating blocks.

    Using the .enter() and .exit() methods on opening and closing
    blocks, we can match up tags of the same level so that later
    it's easier to treat them the same way in the linting engine.

    In case looping means that we encounter the same block more
    than once, we use cache uuids against their source location
    so that if we try to re-enter the block again, it will get
    the same uuid on the second pass.
    """

    _stack: list[UUID] = []
    _map: dict[tuple[int, int], UUID] = {}

    def enter(self, src_slice: slice) -> None:
        """Add a block to the stack."""
        key = to_tuple(src_slice)
        uuid = self._map.get(key, None)

        if not uuid:
            uuid = uuid4()
            self._map[key] = uuid
            lexer_logger.debug(
                "        Entering block stack @ %s: %s (fresh)",
                src_slice,
                uuid,
            )
        else:
            lexer_logger.debug(
                "        Entering block stack @ %s: %s (cached)",
                src_slice,
                uuid,
            )

        self._stack.append(uuid)

    def exit(self) -> None:
        """Pop a block from the stack."""
        uuid = self._stack.pop()
        lexer_logger.debug(
            "        Exiting block stack: %s",
            uuid,
        )

    def top(self) -> UUID:
        """Get the uuid on top of the stack."""
        return self._stack[-1]


class LexedElement(NamedTuple):
    """An element matched during lexing."""

    raw: str
    matcher: "StringLexer"


class TemplateElement(NamedTuple):
    """A LexedElement, bundled with it's position in the templated file."""

    raw: str
    template_slice: slice
    matcher: "StringLexer"

    @classmethod
    def from_element(
        cls, element: LexedElement, template_slice: slice
    ) -> "TemplateElement":
        """Make a TemplateElement from a LexedElement."""
        return cls(
            raw=element.raw, template_slice=template_slice, matcher=element.matcher
        )

    def to_segment(
        self, pos_marker: PositionMarker, subslice: Optional[slice] = None
    ) -> RawSegment:
        """Create a segment from this lexed element."""
        return self.matcher.construct_segment(
            self.raw[subslice] if subslice else self.raw, pos_marker=pos_marker
        )


class LexMatch(NamedTuple):
    """A class to hold matches from the Lexer."""

    forward_string: str
    elements: list[LexedElement]

    def __bool__(self) -> bool:
        """A LexMatch is truthy if it contains a non-zero number of matched elements."""
        return len(self.elements) > 0


LexerType = Union["RegexLexer", "StringLexer"]


class StringLexer:
    """This singleton matcher matches strings exactly.

    This is the simplest usable matcher, but it also defines some of the
    mechanisms for more complicated matchers, which may simply override the
    `_match` function rather than the public `match` function.  This acts as
    the base class for matchers.
    """

    def __init__(
        self,
        name: str,
        template: str,
        segment_class: type[RawSegment],
        subdivider: Optional[LexerType] = None,
        trim_post_subdivide: Optional[LexerType] = None,
        segment_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        self.name = name
        self.template = template
        self.segment_class = segment_class
        self.subdivider = subdivider
        self.trim_post_subdivide = trim_post_subdivide
        self.segment_kwargs = segment_kwargs or {}
        # Guard against the recurring `trim_start="--"` / `trim_chars="'"` typo.
        # These must be tuples/lists of whole sequences to strip: a bare str
        # silently means "strip each of these characters" in Python, and the
        # Rust codegen (utils/build_lexers.py) char-splits it into per-character
        # entries, so the two engines diverge on what a token trims to. Fail
        # loudly at construction instead of shipping that divergence.
        for _key in ("trim_start", "trim_chars"):
            _val = self.segment_kwargs.get(_key)
            if isinstance(_val, str):
                raise TypeError(
                    f"{name}: segment_kwargs[{_key!r}] must be a tuple/list of "
                    f"strings, not a bare str ({_val!r}); write ({_val!r},)."
                )
        self.__post_init__()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name}>"

    def __post_init__(self) -> None:
        """Optional post-init method called after __init__().

        Designed for subclasses to use.
        """
        pass

    def _match(self, forward_string: str) -> Optional[LexedElement]:
        """The private match function. Just look for a literal string."""
        if forward_string.startswith(self.template):
            return LexedElement(self.template, self)
        else:
            return None

    def search(self, forward_string: str) -> Optional[tuple[int, int]]:
        """Use string methods to find a substring."""
        loc = forward_string.find(self.template)
        if loc >= 0:
            return loc, loc + len(self.template)
        else:
            return None

    def _trim_match(self, matched_str: str) -> list[LexedElement]:
        """Given a string, trim if we are allowed to.

        Returns:
            :obj:`tuple` of LexedElement

        """
        elem_buff: list[LexedElement] = []
        content_buff = ""
        str_buff = matched_str

        if self.trim_post_subdivide:
            while str_buff:
                # Iterate through subdividing as appropriate
                trim_pos = self.trim_post_subdivide.search(str_buff)
                # No match? Break
                if not trim_pos:
                    break
                # Start match?
                elif trim_pos[0] == 0:
                    elem_buff.append(
                        LexedElement(
                            str_buff[: trim_pos[1]],
                            self.trim_post_subdivide,
                        )
                    )
                    str_buff = str_buff[trim_pos[1] :]
                # End Match?
                elif trim_pos[1] == len(str_buff):
                    elem_buff += [
                        LexedElement(
                            content_buff + str_buff[: trim_pos[0]],
                            self,
                        ),
                        LexedElement(
                            str_buff[trim_pos[0] : trim_pos[1]],
                            self.trim_post_subdivide,
                        ),
                    ]
                    content_buff, str_buff = "", ""
                # Mid Match? (carry on)
                else:
                    content_buff += str_buff[: trim_pos[1]]
                    str_buff = str_buff[trim_pos[1] :]

        # Do we have anything left? (or did nothing happen)
        if content_buff + str_buff:
            elem_buff.append(
                LexedElement(content_buff + str_buff, self),
            )
        return elem_buff

    def _subdivide(self, matched: LexedElement) -> list[LexedElement]:
        """Given a string, subdivide if we area allowed to.

        Returns:
            :obj:`tuple` of segments

        """
        # Can we have to subdivide?
        if self.subdivider:
            # Yes subdivision
            elem_buff: list[LexedElement] = []
            str_buff = matched.raw
            while str_buff:
                # Iterate through subdividing as appropriate
                div_pos = self.subdivider.search(str_buff)
                if div_pos:
                    # Found a division
                    trimmed_elems = self._trim_match(str_buff[: div_pos[0]])
                    div_elem = LexedElement(
                        str_buff[div_pos[0] : div_pos[1]], self.subdivider
                    )
                    elem_buff += trimmed_elems + [div_elem]
                    str_buff = str_buff[div_pos[1] :]
                else:
                    # No more division matches. Trim?
                    trimmed_elems = self._trim_match(str_buff)
                    elem_buff += trimmed_elems
                    break
            return elem_buff
        else:
            return [matched]

    def match(self, forward_string: str) -> LexMatch:
        """Given a string, match what we can and return the rest.

        Returns:
            :obj:`LexMatch`

        """
        if len(forward_string) == 0:  # pragma: no cover
            raise ValueError("Unexpected empty string!")
        matched = self._match(forward_string)

        if matched:
            # Handle potential subdivision elsewhere.
            new_elements = self._subdivide(matched)

            return LexMatch(
                forward_string[len(matched.raw) :],
                new_elements,
            )
        else:
            return LexMatch(forward_string, [])

    def construct_segment(self, raw: str, pos_marker: PositionMarker) -> RawSegment:
        """Construct a segment using the given class a properties.

        Unless an override `type` is provided in the `segment_kwargs`,
        it is assumed that the `name` of the lexer is designated as the
        intended `type` of the segment.
        """
        # NOTE: Using a private attribute here feels a bit wrong.
        _segment_class_types = self.segment_class._class_types
        _kwargs = self.segment_kwargs
        assert not ("type" in _kwargs and "instance_types" in _kwargs), (
            f"Cannot set both `type` and `instance_types` in segment kwargs: {_kwargs}"
        )
        if "type" in _kwargs:
            # TODO: At some point we should probably deprecate this API and only
            # allow setting `instance_types`.
            assert _kwargs["type"]
            _kwargs["instance_types"] = (_kwargs.pop("type"),)
        elif "instance_types" not in _kwargs and self.name not in _segment_class_types:
            _kwargs["instance_types"] = (self.name,)
        return self.segment_class(raw=raw, pos_marker=pos_marker, **_kwargs)


class RegexLexer(StringLexer):
    """This RegexLexer matches based on regular expressions."""

    def __post_init__(self) -> None:
        """Handle setup for RegexLexer."""
        # We might want to configure this at some point, but for now, newlines
        # do get matched by .
        flags = regex.DOTALL
        self._compiled_regex = regex.compile(self.template, flags)

    def _match(self, forward_string: str) -> Optional[LexedElement]:
        """Use regexes to match chunks."""
        match = self._compiled_regex.match(forward_string)
        if match:
            # We can only match strings with length
            match_str = match.group(0)
            if match_str:
                return LexedElement(match_str, self)
            else:  # pragma: no cover
                lexer_logger.warning(
                    f"Zero length Lex item returned from {self.name!r}. Report this as "
                    "a bug."
                )
        return None

    def search(self, forward_string: str) -> Optional[tuple[int, int]]:
        """Use regex to find a substring."""
        match = self._compiled_regex.search(forward_string)
        if match:
            # We can only match strings with length
            if match.group(0):
                return match.span()
            else:  # pragma: no cover
                lexer_logger.warning(
                    f"Zero length Lex item returned from {self.name!r}. Report this as "
                    "a bug."
                )
        return None


def _handle_zero_length_slice(
    tfs: TemplatedFileSlice,
    next_tfs: Optional[TemplatedFileSlice],
    block_stack: BlockTracker,
    templated_file: TemplatedFile,
    add_indents: bool,
) -> Iterator[MetaSegment]:
    """Generate placeholders and loop segments from a zero length slice.

    This method checks for:
    1. Backward jumps (inserting :obj:`TemplateLoop`).
    2. Forward jumps (inserting :obj:`TemplateSegment`).
    3. Blocks (inserting :obj:`TemplateSegment`).
    4. Unrendered template elements(inserting :obj:`TemplateSegment`).

    For blocks and loops, :obj:`Indent` and :obj:`Dedent` segments are
    yielded around them as appropriate.

    NOTE: block_stack is _mutated_ by this method.
    """
    assert is_zero_slice(tfs.templated_slice)
    # First check for jumps. Backward initially, because in the backward
    # case we don't render the element we find first.
    # That requires being able to look past to the next element.
    if tfs.slice_type.startswith("block") and next_tfs:
        # Look for potential backward jump
        if next_tfs.source_slice.start < tfs.source_slice.start:
            lexer_logger.debug("      Backward jump detected. Inserting Loop Marker")
            # If we're here remember we're on the tfs which is the block end
            # i.e. not the thing we want to render.
            pos_marker = PositionMarker.from_point(
                tfs.source_slice.start,
                tfs.templated_slice.start,
                templated_file,
            )
            if add_indents:
                yield Dedent(
                    is_template=True,
                    pos_marker=pos_marker,
                )

            yield TemplateLoop(pos_marker=pos_marker, block_uuid=block_stack.top())

            if add_indents:
                yield Indent(
                    is_template=True,
                    pos_marker=pos_marker,
                )
            # Move on to the next templated slice. Don't render this directly.
            return

    # Then handle blocks (which aren't jumps backward)
    if tfs.slice_type.startswith("block"):
        # It's a block. Yield a placeholder with potential indents.

        # Update block stack or add indents
        if tfs.slice_type == "block_start":
            block_stack.enter(tfs.source_slice)
        elif add_indents and tfs.slice_type in ("block_end", "block_mid"):
            yield Dedent(
                is_template=True,
                pos_marker=PositionMarker.from_point(
                    tfs.source_slice.start,
                    tfs.templated_slice.start,
                    templated_file,
                ),
                # NOTE: We mark the dedent with the block uuid too.
                block_uuid=block_stack.top(),
            )

        yield TemplateSegment.from_slice(
            tfs.source_slice,
            tfs.templated_slice,
            block_type=tfs.slice_type,
            templated_file=templated_file,
            block_uuid=block_stack.top(),
        )

        # Update block stack or add indents
        if tfs.slice_type == "block_end":
            block_stack.exit()
        elif add_indents and tfs.slice_type in ("block_start", "block_mid"):
            yield Indent(
                is_template=True,
                pos_marker=PositionMarker.from_point(
                    tfs.source_slice.stop,
                    tfs.templated_slice.stop,
                    templated_file,
                ),
                # NOTE: We mark the indent with the block uuid too.
                block_uuid=block_stack.top(),
            )

        # Before we move on, we might have a _forward_ jump to the next
        # element. That element can handle itself, but we'll add a
        # placeholder for it here before we move on.
        if next_tfs and next_tfs.source_slice.start > tfs.source_slice.stop:
            # We do so extract the string.
            placeholder_str = templated_file.source_str[
                tfs.source_slice.stop : next_tfs.source_slice.start
            ]
            # Trim it if it's too long to show.
            if len(placeholder_str) >= 20:
                placeholder_str = (
                    f"... [{len(placeholder_str)} unused template characters] ..."
                )
            lexer_logger.debug("      Forward jump detected. Inserting placeholder")
            yield TemplateSegment(
                pos_marker=PositionMarker(
                    slice(tfs.source_slice.stop, next_tfs.source_slice.start),
                    # Zero slice in the template.
                    tfs.templated_slice,
                    templated_file,
                ),
                source_str=placeholder_str,
                block_type="skipped_source",
            )

        # Move on
        return

    # Always return the slice, even if the source slice was also zero length.  Some
    # templaters might want to pass through totally zero length slices as a way of
    # marking locations in the middle of templated output.
    yield TemplateSegment.from_slice(
        tfs.source_slice,
        tfs.templated_slice,
        tfs.slice_type,
        templated_file,
    )


def _iter_segments(
    lexed_elements: list[TemplateElement],
    templated_file: TemplatedFile,
    add_indents: bool = True,
) -> Iterator[RawSegment]:
    # An index to track where we've got to in the templated file.
    tfs_idx = 0
    # We keep a map of previous block locations in case they re-occur.
    block_stack = BlockTracker()
    templated_file_slices = templated_file.sliced_file

    # Now work out source slices, and add in template placeholders.
    for idx, element in enumerate(lexed_elements):
        # We're working through elements in the rendered file.
        # When they enter this code they don't have a position in the source.
        # We already have a map of how templated elements map to the source file
        # so we work through them to work out what's going on. In theory we can
        # step through the two lists in lock step.

        # i.e. we worked through the lexed elements, but check off the templated
        # file slices as we go.

        # Output the slice as we lex.
        lexer_logger.debug("  %s: %s. [tfs_idx = %s]", idx, element, tfs_idx)

        # All lexed elements, by definition, have a position in the templated
        # file. That means we've potentially got zero-length elements we also
        # need to consider. We certainly need to consider templated slices
        # at tfs_idx. But we should consider some others after that which we
        # might also need to consider.

        # A lexed element is either a literal in the raw file or the result
        # (or part of the result) of a template placeholder. We don't make
        # placeholders for any variables which return a non-zero length of
        # code. We do add placeholders for others.

        # The amount of the current element which has already been consumed.
        consumed_element_length = 0
        # The position in the source which we still need to yield from.
        stashed_source_idx = None

        for tfs_idx, tfs in enumerate(templated_file_slices[tfs_idx:], tfs_idx):
            lexer_logger.debug("      %s: %s", tfs_idx, tfs)

            # Is it a zero slice?
            if is_zero_slice(tfs.templated_slice):
                next_tfs = (
                    templated_file_slices[tfs_idx + 1]
                    if tfs_idx + 1 < len(templated_file_slices)
                    else None
                )
                yield from _handle_zero_length_slice(
                    tfs, next_tfs, block_stack, templated_file, add_indents
                )
                continue

            if tfs.slice_type == "literal":
                # There's a literal to deal with here. Yield as much as we can.

                # Can we cover this whole lexed element with the current templated
                # slice without moving on?
                tfs_offset = tfs.source_slice.start - tfs.templated_slice.start
                # NOTE: Greater than OR EQUAL, to include the case of it matching
                # length exactly.
                if element.template_slice.stop <= tfs.templated_slice.stop:
                    lexer_logger.debug(
                        "     Consuming whole from literal. Existing Consumed: %s",
                        consumed_element_length,
                    )
                    # If we have a stashed start use that. Otherwise infer start.
                    if stashed_source_idx is not None:
                        slice_start = stashed_source_idx
                    else:
                        slice_start = (
                            element.template_slice.start
                            + consumed_element_length
                            + tfs_offset
                        )
                    yield element.to_segment(
                        pos_marker=PositionMarker(
                            slice(
                                slice_start,
                                element.template_slice.stop + tfs_offset,
                            ),
                            element.template_slice,
                            templated_file,
                        ),
                        subslice=slice(consumed_element_length, None),
                    )

                    # If it was an exact match, consume the templated element too.
                    if element.template_slice.stop == tfs.templated_slice.stop:
                        tfs_idx += 1
                    # In any case, we're done with this element. Move on
                    break
                elif element.template_slice.start == tfs.templated_slice.stop:
                    # Did we forget to move on from the last tfs and there's
                    # overlap?
                    # NOTE: If the rest of the logic works, this should never
                    # happen.
                    lexer_logger.debug("     NOTE: Missed Skip")  # pragma: no cover
                    continue  # pragma: no cover
                else:
                    # This means that the current lexed element spans across
                    # multiple templated file slices.
                    lexer_logger.debug("     Consuming whole spanning literal")
                    # This almost certainly means there's a templated element
                    # in the middle of a whole lexed element.

                    # What we do here depends on whether we're allowed to split
                    # lexed elements. This is basically only true if it's whitespace.
                    # NOTE: We should probably make this configurable on the
                    # matcher object, but for now we're going to look for the
                    # name of the lexer.
                    if element.matcher.name == "whitespace":
                        # We *can* split it!
                        # Consume what we can from this slice and move on.
                        lexer_logger.debug(
                            "     Consuming split whitespace from literal. "
                            "Existing Consumed: %s",
                            consumed_element_length,
                        )
                        if stashed_source_idx is not None:
                            raise NotImplementedError(  # pragma: no cover
                                "Found literal whitespace with stashed idx!"
                            )
                        incremental_length = (
                            tfs.templated_slice.stop - element.template_slice.start
                        )
                        yield element.to_segment(
                            pos_marker=PositionMarker(
                                slice(
                                    element.template_slice.start
                                    + consumed_element_length
                                    + tfs_offset,
                                    tfs.templated_slice.stop + tfs_offset,
                                ),
                                element.template_slice,
                                templated_file,
                            ),
                            # Subdivide the existing segment.
                            subslice=offset_slice(
                                consumed_element_length,
                                incremental_length,
                            ),
                        )
                        consumed_element_length += incremental_length
                        continue
                    else:
                        # We can't split it. We're going to end up yielding a segment
                        # which spans multiple slices. Stash the type, and if we haven't
                        # set the start yet, stash it too.
                        lexer_logger.debug("     Spilling over literal slice.")
                        if stashed_source_idx is None:
                            stashed_source_idx = (
                                element.template_slice.start + tfs_offset
                            )
                            lexer_logger.debug(
                                "     Stashing a source start. %s", stashed_source_idx
                            )
                        continue

            elif tfs.slice_type in ("templated", "block_start", "escaped"):
                # Found a templated slice. Does it have length in the templated file?
                # If it doesn't, then we'll pick it up next.
                if not is_zero_slice(tfs.templated_slice):
                    # If it's a block_start. Append to the block stack.
                    # NOTE: This is rare, but call blocks do occasionally
                    # have length (and so don't get picked up by
                    # _handle_zero_length_slice)
                    if tfs.slice_type == "block_start":
                        block_stack.enter(tfs.source_slice)

                    # Is our current element totally contained in this slice?
                    if element.template_slice.stop <= tfs.templated_slice.stop:
                        lexer_logger.debug("     Contained templated slice.")
                        # Yes it is. Add lexed element with source slices as the whole
                        # span of the source slice for the file slice.
                        # If we've got an existing stashed source start, use that
                        # as the start of the source slice.
                        if stashed_source_idx is not None:
                            slice_start = stashed_source_idx
                        else:
                            slice_start = (
                                tfs.source_slice.start + consumed_element_length
                            )
                        yield element.to_segment(
                            pos_marker=PositionMarker(
                                slice(
                                    slice_start,
                                    # The end in the source is the end of the templated
                                    # slice. We can't subdivide any better.
                                    tfs.source_slice.stop,
                                ),
                                element.template_slice,
                                templated_file,
                            ),
                            subslice=slice(consumed_element_length, None),
                        )

                        # If it was an exact match, consume the templated element too.
                        if element.template_slice.stop == tfs.templated_slice.stop:
                            tfs_idx += 1
                        # Carry on to the next lexed element
                        break
                    # We've got an element which extends beyond this templated slice.
                    # This means that a _single_ lexed element claims both some
                    # templated elements and some non-templated elements. That could
                    # include all kinds of things (and from here we don't know what
                    # else is yet to come, comments, blocks, literals etc...).

                    # In the `literal` version of this code we would consider
                    # splitting the literal element here, but in the templated
                    # side we don't. That's because the way that templated tokens
                    # are lexed, means that they should arrive "pre-split".
                    else:
                        # Stash the source idx for later when we do make a segment.
                        lexer_logger.debug("     Spilling over templated slice.")
                        if stashed_source_idx is None:
                            stashed_source_idx = tfs.source_slice.start
                            lexer_logger.debug(
                                "     Stashing a source start as lexed element spans "
                                "over the end of a template slice. %s",
                                stashed_source_idx,
                            )
                        # Move on to the next template slice
                        continue

            raise NotImplementedError(
                f"Unable to process slice: {tfs}"
            )  # pragma: no cover

    # If templated elements are left, yield them.
    # We can assume they're all zero length if we're here.
    for tfs_idx, tfs in enumerate(templated_file_slices[tfs_idx:], tfs_idx):
        next_tfs = (
            templated_file_slices[tfs_idx + 1]
            if tfs_idx + 1 < len(templated_file_slices)
            else None
        )
        yield from _handle_zero_length_slice(
            tfs, next_tfs, block_stack, templated_file, add_indents
        )


class PyLexer:
    """The Lexer class actually does the lexing step."""

    def __init__(
        self,
        config: Optional[FluffConfig] = None,
        last_resort_lexer: Optional[StringLexer] = None,
        dialect: Optional[str] = None,
    ):
        if config and dialect:
            raise ValueError(  # pragma: no cover
                "Lexer does not support setting both `config` and `dialect`."
            )
        # Use the provided config or create one from the dialect.
        self.config = config or FluffConfig.from_kwargs(dialect=dialect)
        # Store the matchers
        self.lexer_matchers = self.config.get("dialect_obj").get_lexer_matchers()

        self.last_resort_lexer = last_resort_lexer or RegexLexer(
            "<unlexable>",
            r"[^\t\n\ ]*",
            UnlexableSegment,
        )

    def lex(
        self, raw: Union[str, TemplatedFile]
    ) -> tuple[tuple[BaseSegment, ...], list[SQLLexError]]:
        """Take a string or TemplatedFile and return segments.

        If we fail to match the *whole* string, then we must have
        found something that we cannot lex. If that happens we should
        package it up as unlexable and keep track of the exceptions.
        """
        lexer_logger.info("Lexing file using PyLexer.")
        # Make sure we've got a string buffer and a template
        # regardless of what was passed in.
        if isinstance(raw, str):
            template = TemplatedFile.from_string(raw)
            str_buff = raw
        else:
            template = raw
            str_buff = str(template)

        # Lex the string to get a tuple of LexedElement
        element_buffer: list[LexedElement] = []
        while True:
            res = self.lex_match(str_buff, self.lexer_matchers)
            element_buffer += res.elements
            if res.forward_string:
                resort_res = self.last_resort_lexer.match(res.forward_string)
                if not resort_res:  # pragma: no cover
                    # If we STILL can't match, then just panic out.
                    raise SQLLexError(
                        "Fatal. Unable to lex characters: {0!r}".format(
                            res.forward_string[:10] + "..."
                            if len(res.forward_string) > 9
                            else res.forward_string
                        )
                    )
                str_buff = resort_res.forward_string
                element_buffer += resort_res.elements
            else:  # pragma: no cover TODO?
                break

        # Map tuple LexedElement to list of TemplateElement.
        # This adds the template_slice to the object.
        templated_buffer = self.map_template_slices(element_buffer, template)

        # Turn lexed elements into segments.
        segments: tuple[RawSegment, ...] = self.elements_to_segments(
            templated_buffer, template
        )

        # Generate any violations
        violations: list[SQLLexError] = self.violations_from_segments(segments)

        return segments, violations

    def elements_to_segments(
        self, elements: list[TemplateElement], templated_file: TemplatedFile
    ) -> tuple[RawSegment, ...]:
        """Convert a tuple of lexed elements into a tuple of segments."""
        lexer_logger.info("Elements to Segments.")
        add_indents = self.config.get("template_blocks_indent", "indentation")
        # Delegate to _iter_segments
        segment_buffer: list[RawSegment] = list(
            _iter_segments(elements, templated_file, add_indents)
        )

        # Add an end of file marker
        segment_buffer.append(
            EndOfFile(
                pos_marker=(
                    segment_buffer[-1].pos_marker.end_point_marker()
                    if segment_buffer
                    else PositionMarker.from_point(0, 0, templated_file)
                )
            )
        )
        # Convert to tuple before return
        return tuple(segment_buffer)

    @staticmethod
    def violations_from_segments(
        segments: tuple[RawSegment, ...],
    ) -> list[SQLLexError]:
        """Generate any lexing errors for any unlexables."""
        violations = []
        for segment in segments:
            if segment.is_type("unlexable"):
                violations.append(
                    SQLLexError(
                        "Unable to lex characters: {!r}".format(
                            segment.raw[:10] + "..."
                            if len(segment.raw) > 9
                            else segment.raw
                        ),
                        pos=segment.pos_marker,
                    )
                )
        return violations

    @staticmethod
    def lex_match(forward_string: str, lexer_matchers: list[StringLexer]) -> LexMatch:
        """Iteratively match strings using the selection of submatchers."""
        elem_buff: list[LexedElement] = []
        while True:
            if len(forward_string) == 0:
                return LexMatch(forward_string, elem_buff)
            for matcher in lexer_matchers:
                res = matcher.match(forward_string)
                if res.elements:
                    # If we have new segments then whoop!
                    elem_buff += res.elements
                    forward_string = res.forward_string
                    # Cycle back around again and start with the top
                    # matcher again.
                    break
            else:
                # We've got so far, but now can't match. Return
                return LexMatch(forward_string, elem_buff)

    @staticmethod
    def map_template_slices(
        elements: list[LexedElement], template: TemplatedFile
    ) -> list[TemplateElement]:
        """Create a tuple of TemplateElement from a tuple of LexedElement.

        This adds slices in the templated file to the original lexed
        elements. We'll need this to work out the position in the source
        file.
        """
        idx = 0
        templated_buff: list[TemplateElement] = []
        for element in elements:
            template_slice = offset_slice(idx, len(element.raw))
            idx += len(element.raw)
            templated_buff.append(TemplateElement.from_element(element, template_slice))
            if (
                template.templated_str[template_slice] != element.raw
            ):  # pragma: no cover
                raise ValueError(
                    "Template and lexed elements do not match. This should never "
                    f"happen {element.raw!r} != "
                    f"{template.templated_str[template_slice]!r}"
                )
        return templated_buff


try:
    from sqlfluffrs import RsLexer, RsToken

    def get_segment_type_map(base_class: type) -> dict[str, type[RawSegment]]:
        """Dynamically create a map of segment types to their subclasses."""
        segment_map = {}
        for subclass in base_class.__subclasses__():
            if subclass is LiteralKeywordSegment or subclass is ImplicitIndent:
                continue
            if (
                hasattr(subclass, "type") and subclass.type
            ):  # Ensure the subclass has a type
                segment_map[subclass.type] = subclass
            # Recursively add subclasses of subclasses
            segment_map.update(get_segment_type_map(subclass))
        return segment_map

    # Dynamically generate the segment_types map
    segment_types = get_segment_type_map(RawSegment)

    class PyRsLexer(RsLexer):
        """A wrapper around the sqlfluffrs lexer."""

        @staticmethod
        def _tokens_to_segments(
            tokens: list["RsToken"], py_template: TemplatedFile
        ) -> tuple[BaseSegment, ...]:
            """Convert tokens to segments."""
            return tuple(
                segment_types.get(token.type, RawSegment).from_rstoken(
                    token, py_template
                )
                for token in tokens
            )

        def lex(
            self, raw: Union[str, TemplatedFile]
        ) -> tuple[tuple[BaseSegment, ...], list[SQLLexError]]:
            """Take a string or TemplatedFile and return segments."""
            lexer_logger.info("Lexing file using RsLexer.")
            tokens, errors = self._lex(raw)
            first_token = tokens[0]
            assert first_token
            template = first_token.pos_marker.templated_file
            py_template = TemplatedFile(
                template.source_str,
                template.fname,
                template.templated_str,
                template.sliced_file,  # type: ignore
                template.raw_sliced,  # type: ignore
            )

            return (
                self._tokens_to_segments(tokens, py_template),
                [SQLLexError.from_rs_error(error) for error in errors],
            )

    _HAS_RUST_LEXER = True
except ImportError:
    PyRsLexer = None  # type: ignore[assignment, misc]
    _HAS_RUST_LEXER = False


def get_lexer_class() -> type[Union[PyLexer, "PyRsLexer"]]:
    """Get the appropriate lexer class based on availability.

    Returns PyRsLexer if the Rust extension is available,
    otherwise returns PyLexer.

    This function provides a single point of lexer selection,
    making it easy to instantiate the correct lexer:

        Lexer = get_lexer_class()
        lexer = Lexer(config=config)

    Returns:
        The lexer class to use (PyRsLexer or PyLexer).
    """
    if _HAS_RUST_LEXER:
        lexer_logger.info("Rust extensions are available. Using PyRsLexer.")
        return PyRsLexer
    lexer_logger.info("Rust extensions are not available. Using PyLexer.")
    return PyLexer



from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union
from uuid import UUID

if TYPE_CHECKING:
    from sqlfluff.core.config import FluffConfig
    from sqlfluff.core.parser.lexer import StringLexer
    from sqlfluff.core.parser.segments import SourceFix
    from sqlfluff.core.templaters import TemplatedFile

SerializedObject = dict[str, Union[str, int, bool, list["SerializedObject"]]]
TupleSerialisedSegment = tuple[str, Union[str, tuple["TupleSerialisedSegment", ...]]]

class Slice: ...

class RsRawFileSlice:
    raw: str
    slice_type: str
    source_idx: int
    block_idx: int
    tag: Optional[str]

class RsTemplatedFileSlice:
    slice_type: str
    source_slice: Slice
    templated_slice: Slice

class RsTemplatedFile:
    source_str: str
    fname: str
    templated_str: str
    sliced_file: List[RsTemplatedFileSlice]
    raw_sliced: List[RsRawFileSlice]

class RsPositionMarker:
    source_slice: slice
    templated_slice: slice
    templated_file: RsTemplatedFile
    working_line_no: int
    working_line_pos: int

class RsToken:
    raw: str
    pos_marker: RsPositionMarker
    type: str
    uuid: Optional[int]
    source_fixes: Optional[list["SourceFix"]]

    def raw_trimmed(self) -> str: ...
    @property
    def is_templated(self) -> bool: ...
    @property
    def is_code(self) -> bool: ...
    @property
    def is_meta(self) -> bool: ...
    @property
    def source_str(self) -> str: ...
    @property
    def block_type(self) -> str: ...
    @property
    def block_uuid(self) -> Optional[UUID]: ...
    @property
    def cache_key(self) -> str: ...
    @property
    def trim_start(self) -> Optional[tuple[str]]: ...
    @property
    def trim_chars(self) -> Optional[tuple[str]]: ...
    @property
    def quoted_value(self) -> Optional[tuple[str, int | str]]: ...
    @property
    def escape_replacements(self) -> Optional[list[tuple[str, str]]]: ...
    def count_segments(self, raw_only: bool = False) -> int: ...
    def get_type(self) -> str: ...
    def recursive_crawl(
        self,
        seg_type: Tuple[str, ...],
        recurse_into: bool,
        no_recursive_seg_type: Optional[Union[str, List[str]]] = None,
        allow_self: bool = True,
    ) -> List["RsToken"]: ...
    def recursive_crawl_all(self, reverse: bool) -> List["RsToken"]: ...
    @property
    def segments(self) -> List["RsToken"]: ...
    def path_to(self, other: "RsToken") -> List[Any]: ...
    def get_start_loc(self) -> Tuple[int, int]: ...
    def get_end_loc(self) -> Tuple[int, int]: ...
    @property
    def raw_segments(self) -> List["RsToken"]: ...
    def copy(
        self,
        segments: Optional[List["RsToken"]] = None,
        parent: Optional[Any] = None,
        parent_idx: Optional[int] = None,
    ) -> "RsToken": ...
    def edit(
        self,
        raw: Optional[str] = None,
        source_fixes: Optional[List[Any]] = None,
    ) -> "RsToken": ...
    def to_tuple(
        self,
        code_only: Optional[bool] = None,
        show_raw: Optional[bool] = None,
        include_meta: Optional[bool] = None,
    ) -> TupleSerialisedSegment: ...
    def __repr__(self) -> str: ...
    @property
    def instance_types(self) -> List[str]: ...
    @staticmethod
    def template_placeholder_from_slice(
        source_slice: tuple[int, int],
        templated_slice: tuple[int, int],
        block_type: str,
        _source_str: str,
        block_uuid: Optional[str],
        templated_file: "TemplatedFile",
    ) -> "RsToken": ...

class RsNode: ...

class RsSQLLexerError:
    desc: str
    line_no: int
    line_pos: int
    ignore: bool
    warning: bool
    fatal: bool

    def __init__(
        self,
        msg: Optional[str] = None,
        pos: Optional[RsPositionMarker] = None,
        line_no: int = 0,
        line_pos: int = 0,
        ignore: bool = False,
        warning: bool = False,
        fatal: bool = False,
    ) -> None: ...
    def rule_code(self) -> str: ...
    def rule_name(self) -> str: ...
    def source_signature(self) -> Tuple[Tuple[str, int, int], str]: ...
    def to_dict(self) -> SerializedObject: ...
    def ignore_if_in(self, ignore_iterable: list[str]) -> None: ...
    def warning_if_in(self, ignore_iterable: list[str]) -> None: ...

class RsLexer:
    def __init__(
        self,
        config: Optional["FluffConfig"] = None,
        last_resort_lexer: Optional["StringLexer"] = None,
        dialect: Optional[str] = None,
    ): ...
    def _lex(
        self, lex_input: Union[str, "TemplatedFile"]
    ) -> Tuple[List[RsToken], List[Any]]: ...

class RsMatchResult:
    """Result of a Rust parser match operation."""

    matched_slice: tuple[int, int]
    matched_class: Optional[str]
    child_matches: List["RsMatchResult"]
    parse_error: Optional[tuple[str, int]]
    instance_types: Optional[List[str]]
    segment_kwargs: Optional[dict[str, Any]]
    trim_chars: Optional[List[str]]
    casefold: Optional[str]
    quoted_value: Optional[str]
    escape_replacements: Optional[List[tuple[str, str]]]
    insert_segments: Optional[List[tuple[int, str, bool]]]

    def apply_as_tree(
        self,
        tokens: List[RsToken],
        leading: List[RsToken],
        trailing: List[RsToken],
    ) -> "RsTree": ...

class RsHandle:
    """A lightweight cursor into an :class:`RsTree` arena node.

    Every accessor runs Rust-side; only thin handles and scalars cross FFI.
    Wrapped by the Python ``RsSegment`` facade.
    """

    uuid: int
    raw: str
    raw_upper: str
    type: str
    segment_class: Optional[str]
    is_code: bool
    is_whitespace: bool
    is_comment: bool
    is_meta: bool
    is_templated: bool
    pos_marker: Optional[RsPositionMarker]
    children: List["RsHandle"]
    parent: Optional["RsHandle"]

    def is_type(self, seg_type: List[str]) -> bool: ...
    def is_raw(self) -> bool: ...
    def class_types(self) -> List[str]: ...
    def instance_types(self) -> List[str]: ...
    def is_implicit(self) -> Optional[bool]: ...
    def trim_chars(self) -> Optional[List[str]]: ...
    def quoted_value(self) -> Optional[tuple[str, str]]: ...
    def escape_replacements(self) -> Optional[List[tuple[str, str]]]: ...
    def descendant_type_set(self) -> List[str]: ...
    def get_parent(self) -> Optional[tuple["RsHandle", int]]: ...
    def get_child(self, seg_type: List[str]) -> Optional["RsHandle"]: ...
    def get_children(self, seg_type: List[str]) -> List["RsHandle"]: ...
    def raw_segments(self) -> List["RsHandle"]: ...
    def recursive_crawl(
        self,
        seg_type: List[str],
        recurse_into: bool = True,
        no_recursive_seg_type: List[str] = ...,
        allow_self: bool = True,
    ) -> List["RsHandle"]: ...
    def recursive_crawl_all(self) -> List["RsHandle"]: ...
    def path_to(
        self, other: "RsHandle"
    ) -> List[tuple["RsHandle", int, int, List[int]]]: ...

class RsTree:
    """Owner of a mutable arena parse tree, used by the ``RsSegment`` facade."""

    root: RsHandle

    def __len__(self) -> int: ...
    def node_by_uuid(self, uuid: int) -> Optional[RsHandle]: ...

class RsParseError(Exception):
    """Exception raised by Rust parser when parsing fails.

    Attributes:
        pos: Position index in the segments array where the error occurred
    """

    pos: int

class RsParser:
    """Rust-based SQL parser."""

    def __init__(
        self,
        dialect: str,
        indent_config: Optional[dict[str, bool]] = None,
        max_parser_iterations: Optional[int] = None,
        parser_warn_threshold: Optional[int] = None,
        max_parse_depth: int = 0,
        max_parse_nodes: int = 0,
    ): ...
    def parse_match_result_from_tokens(
        self, tokens: List[RsToken]
    ) -> RsMatchResult: ...

def cp01_violations(
    tree: RsTree,
    policy: str,
    ignore_words: List[str] = ...,
    ignore_templated: bool = ...,
) -> List[Tuple[int, str]]:
    """Detect CP01 (keyword capitalisation) violations natively over the arena.

    Returns ``(leaf_index, fixed_raw)`` pairs; ``leaf_index`` is 1:1 with the
    parse tree's ``raw_segments`` order.
    """
    ...

def cp03_violations(
    tree: RsTree,
    policy: str,
    ignore_words: List[str] = ...,
    ignore_templated: bool = ...,
) -> List[Tuple[int, str]]:
    """Detect CP03 (function-name capitalisation) violations natively over the arena.

    Same contract as ``cp01_violations``; ``policy`` additionally accepts
    ``pascal``/``camel``/``snake``.
    """
    ...

def cp04_violations(
    tree: RsTree,
    policy: str,
    ignore_words: List[str] = ...,
    ignore_templated: bool = ...,
) -> List[Tuple[int, str]]:
    """Detect CP04 (boolean/null literal capitalisation) violations natively over the arena.

    Same contract as ``cp01_violations``.
    """
    ...



"""Implementation of Rule CV12."""

import collections
from collections.abc import Iterator
from typing import Deque

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.parser.segments.common import (
    BinaryOperatorSegment,
    WhitespaceSegment,
)
from sqlfluff.core.parser.segments.keyword import KeywordSegment
from sqlfluff.core.rules import BaseRule, EvalResultType, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.fix import LintFix
from sqlfluff.dialects.dialect_ansi import (
    ExpressionSegment,
    JoinClauseSegment,
    JoinOnConditionSegment,
)


class Rule_CV12(BaseRule):
    """Use `JOIN ... ON ...` instead of `WHERE ...` for join conditions.

    **Anti-pattern**

    Using WHERE clause for join conditions.

    .. code-block:: sql

        SELECT
            foo.a
            , bar.b
        FROM foo
        JOIN bar
        WHERE foo.x = bar.y;

    **Best practice**

    Use JOIN ON clause for join condition.

    .. code-block:: sql

        SELECT
            foo.a
            , bar.b
        FROM foo
        JOIN bar
        ON foo.x = bar.y;

    """

    name = "convention.join_condition"
    aliases = ()
    groups = ("all", "convention")
    crawl_behaviour = SegmentSeekerCrawler({"select_statement"})
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> EvalResultType:
        """Find joins with WHERE clause.

        Fix them into JOIN ON.
        """
        return [lint_result for lint_result in self._eval_gen(context)]

    def _eval_gen(self, context: RuleContext) -> Iterator[LintResult]:
        # We are only interested in SELECT statement.
        select_statement = context.segment
        assert select_statement.is_type("select_statement")

        maybe_where_clause = select_statement.get_child("where_clause")
        if not maybe_where_clause:
            return

        where_clause = maybe_where_clause
        where_clause_simplifable = self._is_where_clause_simplifable(where_clause)

        if where_clause_simplifable:
            expr = where_clause.get_child("expression")
            assert expr is not None
            subexpressions = self._get_subexpression_chunks(expr)
        else:
            subexpressions = []
        consumed_subexpressions = set()

        # get references in from clause
        select_table_references = [
            *select_statement.recursive_crawl(
                "from_expression_element",
                no_recursive_seg_type=["join_clause", "select_statement"],
            )
        ]

        # track all seen references (from clause + all previous joins)
        encountered_references = {
            self._get_from_expression_element_alias(table_ref)
            for table_ref in select_table_references
        }

        for join_clause in select_statement.recursive_crawl(
            "join_clause", no_recursive_seg_type=["select_statement"]
        ):
            # mark table reference as seen
            join_table_reference = next(
                join_clause.recursive_crawl(
                    "from_expression_element",
                    no_recursive_seg_type=["select_statement"],
                )
            )
            encountered_function = self._get_from_expression_element_function(
                join_table_reference
            )
            if encountered_function in ("UNNEST",):
                # If UNNEST function is used, disregard lack of condition
                continue

            this_join_reference = self._get_from_expression_element_alias(
                join_table_reference
            )
            encountered_references.add(this_join_reference)

            join_clause_keywords = [
                seg for seg in join_clause.segments if seg.type == "keyword"
            ]

            if any(
                kw.raw_upper in ("CROSS", "NATURAL", "POSITIONAL", "USING", "APPLY")
                for kw in join_clause_keywords
            ):
                # If explicit CROSS JOIN is used, disregard lack of condition
                # If NATURAL JOIN is used, disregard lack of condition
                # If explicit POSITIONAL JOIN is used, disregard lack of condition
                # If explicit JOIN USING is used, disregard lack of condition
                # If explicit CROSS/OUTER APPLY is used, disregard lack of condition
                continue

            this_join_condition = join_clause.get_child("join_on_condition")
            if this_join_condition:
                # Join condition is present, no error reported.
                continue

            if join_clause.is_templated:
                # Moving a condition into an ON clause rewrites the whole join
                # clause. If that clause contains templated code, the resulting
                # patch can't be mapped back onto the source file and is
                # silently dropped - but the WHERE clause rewrite below is
                # literal, so it still applies. The net effect is that the join
                # condition is deleted rather than moved, silently changing the
                # meaning of the query. Flag the violation, but don't offer a
                # fix we can't apply safely.
                yield LintResult(anchor=join_clause)
                continue

            if not where_clause_simplifable:
                yield LintResult(anchor=join_clause)
            else:
                this_join_clause_subexpressions = set()
                for subexpr_idx, subexpr_segments in enumerate(subexpressions):
                    if subexpr_idx in consumed_subexpressions:
                        continue
                    qualified_column_references = [
                        col_ref
                        for seg in subexpr_segments
                        for col_ref in seg.recursive_crawl(
                            "column_reference",
                            no_recursive_seg_type="select_statement",
                        )
                        if "dot" in col_ref.descendant_type_set
                    ]
                    if (
                        len(qualified_column_references) > 1
                        and all(
                            col_ref.raw_upper.startswith(
                                tuple(
                                    f"{table_ref}."
                                    for table_ref in encountered_references
                                )
                            )
                            for col_ref in qualified_column_references
                        )
                        # A condition that doesn't reference this join's own
                        # table belongs to an earlier join, and moving it here
                        # would change the query's semantics.
                        and any(
                            col_ref.raw_upper.startswith(f"{this_join_reference}.")
                            for col_ref in qualified_column_references
                        )
                    ):
                        this_join_clause_subexpressions.add(subexpr_idx)
                        consumed_subexpressions.add(subexpr_idx)

                if not this_join_clause_subexpressions:
                    yield LintResult(join_clause)
                else:
                    join_clause_fix_segments: Deque[BaseSegment] = collections.deque()
                    for subexpr_idx, subexpr_segments in enumerate(subexpressions):
                        if subexpr_idx in this_join_clause_subexpressions:
                            join_clause_fix_segments.extend(subexpr_segments)
                            join_clause_fix_segments.append(
                                BinaryOperatorSegment("AND")
                            )

                    while join_clause_fix_segments and join_clause_fix_segments[
                        0
                    ].is_type("whitespace", "binary_operator"):
                        join_clause_fix_segments.popleft()
                    while join_clause_fix_segments and join_clause_fix_segments[
                        -1
                    ].is_type("whitespace", "binary_operator"):
                        join_clause_fix_segments.pop()

                    join_on_expression = ExpressionSegment(
                        tuple(join_clause_fix_segments),
                    )
                    join_on = JoinOnConditionSegment(
                        (
                            KeywordSegment("ON"),
                            WhitespaceSegment(),
                            join_on_expression,
                        )
                    )
                    join_clause_segment = JoinClauseSegment(
                        (
                            *join_clause.segments,
                            WhitespaceSegment(),
                            join_on,
                        )
                    )

                    yield LintResult(
                        anchor=join_clause,
                        fixes=[
                            LintFix.replace(
                                join_clause,
                                edit_segments=[join_clause_segment],
                            )
                        ],
                    )

        if not where_clause_simplifable:
            return

        if not consumed_subexpressions:
            return

        # Rewrite WHERE to keep conditions not moved to ON clauses
        where_clause_fix_segments: Deque[BaseSegment] = collections.deque()
        for subexpr_idx, subexpr_segments in enumerate(subexpressions):
            if subexpr_idx not in consumed_subexpressions:
                where_clause_fix_segments.extend(subexpr_segments)
                where_clause_fix_segments.append(BinaryOperatorSegment("AND"))

        while where_clause_fix_segments and where_clause_fix_segments[0].is_type(
            "whitespace", "binary_operator"
        ):
            where_clause_fix_segments.popleft()
        while where_clause_fix_segments and where_clause_fix_segments[-1].is_type(
            "whitespace", "binary_operator"
        ):
            where_clause_fix_segments.pop()

        if where_clause_fix_segments:
            where_clause_expr = where_clause.get_child("expression")
            assert where_clause_expr is not None
            yield LintResult(
                anchor=where_clause_expr,
                fixes=[
                    LintFix.replace(
                        where_clause_expr, edit_segments=[*where_clause_fix_segments]
                    )
                ],
            )
        else:
            assert select_statement.segments[-1].is_type("where_clause")
            assert select_statement.segments[-2].is_type("whitespace", "newline")
            yield LintResult(
                anchor=where_clause,
                fixes=[
                    LintFix.delete(select_statement.segments[-2]),
                    LintFix.delete(select_statement.segments[-1]),
                ],
            )

    @staticmethod
    def _get_from_expression_element_alias(from_expr_element: BaseSegment) -> str:
        if "alias_expression" in from_expr_element.direct_descendant_type_set:
            alias_seg = from_expr_element.get_child("alias_expression")
            assert alias_seg is not None
            identifier_seg = alias_seg.get_child("identifier")
            assert identifier_seg is not None
            alias_str = identifier_seg.raw_upper
        else:
            alias_str = from_expr_element.raw_upper

        return alias_str

    @staticmethod
    def _get_from_expression_element_function(from_expr_element: BaseSegment) -> str:
        """Extract the function name from a from_expression_element if present."""

        def get_segment_or_default(
            segment: BaseSegment, child_type: str, default: BaseSegment
        ) -> BaseSegment:
            """Helper to safely get a child segment or return a default."""
            child = segment.get_child(child_type)
            return child if child else default

        table_expr = get_segment_or_default(
            from_expr_element, "table_expression", from_expr_element
        )

        if "function" not in table_expr.direct_descendant_type_set:
            return ""

        function = get_segment_or_default(table_expr, "function", table_expr)
        function_name = get_segment_or_default(function, "function_name", table_expr)
        function_id = get_segment_or_default(
            function_name, "function_name_identifier", table_expr
        )

        return function_id.raw_upper

    @staticmethod
    def _is_where_clause_simplifable(where_clause: BaseSegment) -> bool:
        assert where_clause.is_type("where_clause")
        expr = where_clause.get_child("expression")
        if not expr:  # pragma: no cover
            # According to grammar, we should always have an ExpressionSegment
            # See sqlfluff.dialects.dialect_ansi.WhereClauseSegment
            return False
        ops = expr.recursive_crawl("binary_operator")
        return all(op.raw_upper == "AND" for op in ops)

    @staticmethod
    def _get_subexpression_chunks(expr: BaseSegment) -> list[list[BaseSegment]]:
        expr_segments = expr.segments
        bin_op_indices = [
            i for i, e in enumerate(expr_segments) if e.is_type("binary_operator")
        ]
        split_segments = [None, *[expr_segments[i] for i in bin_op_indices], None]
        start_segments_iter = iter(split_segments)
        stop_segments_iter = iter(split_segments)
        _ = next(stop_segments_iter)
        return [
            expr.select_children(start_seg, stop_seg)
            for start_seg, stop_seg in zip(start_segments_iter, stop_segments_iter)
        ]



"""Implementation of Rule AL10."""

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import BaseRule, EvalResultType, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_AL10(BaseRule):
    """Derived tables must have an alias.

    A derived table (subquery in a ``FROM`` clause) without an alias will
    cause a syntax error in most SQL dialects including MySQL, PostgreSQL,
    and T-SQL.

    **Anti-pattern**

    A subquery in a ``FROM`` clause without an alias.

    .. code-block:: sql

        SELECT *
        FROM (
            SELECT 1 AS a
        )

    **Best practice**

    Add an alias to the derived table.

    .. code-block:: sql

        SELECT *
        FROM (
            SELECT 1 AS a
        ) AS derived

    """

    name = "aliasing.required"
    aliases = ()
    groups = ("all", "core", "aliasing")
    crawl_behaviour = SegmentSeekerCrawler({"from_expression_element"})
    is_fix_compatible = False

    def _eval(self, context: RuleContext) -> EvalResultType:
        """Check that derived tables have an alias."""
        assert context.segment.is_type("from_expression_element")

        # Check if this FROM expression contains a derived table.
        if not self._contains_derived_table(context.segment):
            return None

        # It's a derived table. Check if it has an alias.
        if context.segment.get_child("alias_expression"):
            return None

        # Derived table without an alias.
        return LintResult(
            anchor=context.segment,
            description="Derived table must have an alias.",
        )

    @staticmethod
    def _contains_derived_table(from_expression_element: BaseSegment) -> bool:
        """Check whether a FROM expression element contains a derived table.

        A derived table is a subquery (SELECT, set expression, or WITH-compound
        statement) nested inside a FROM clause, potentially wrapped in brackets.
        """
        for segment in from_expression_element.iter_segments(expanding=("bracketed",)):
            if segment.is_type("table_expression"):
                # Check for nested SELECT, UNION/INTERSECT/EXCEPT, or
                # WITH-compound statement.
                for seg in segment.iter_segments(expanding=("bracketed",)):
                    if seg.is_type(
                        "select_statement",
                        "set_expression",
                        "with_compound_statement",
                    ):
                        return True
        return False
